import contextlib
from dataclasses import asdict
from datetime import datetime
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    KeysView,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from feast.data_source import DataSource
from feast.errors import InvalidEntityType
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.registry.registry import Registry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import RepoConfig
from feast.usage import log_exceptions_and_usage
from feast.infra.offline_stores.contrib.doris_offline_store.doris_source import DorisSQLSource
from feast.infra.utils.doris.connection_util import (
    _get_conn,
    df_to_doris_table,
    get_query_schema,
)
from feast.infra.utils.doris.doris_config import (
    DorisSQLConfig,
    doris_type_code_to_arrow,
)
from jinja2 import BaseLoader, Environment
from pytz import utc


class DorisSQLOfflineStoreConfig(DorisSQLConfig):
    type: Literal["doris"] = "doris"


class DorisSQLOfflineStore(OfflineStore):
    @staticmethod
    @log_exceptions_and_usage(offline_store="doris")
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, DorisSQLOfflineStoreConfig)
        assert isinstance(data_source, DorisSQLSource)
        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(_append_alias(join_key_columns, "a"))
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [timestamp_field]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(_append_alias(timestamps, "a")) + " DESC"
        a_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "a")
        )
        b_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "b")
        )

        query = f"""
            SELECT
                {b_field_string}
                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
            FROM (
                SELECT {a_field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression} a
                WHERE a.{timestamp_field} BETWEEN '{start_date}' AND '{end_date}'
            ) b
            WHERE _feast_row = 1
            """

        return DorisSQLRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="doris")
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: Registry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, DorisSQLOfflineStoreConfig)
        for fv in feature_views:
            assert isinstance(fv.batch_source, DorisSQLSource)

        entity_schema = _get_entity_schema(entity_df, config)

        entity_df_event_timestamp_col = (
            offline_utils.infer_event_timestamp_from_entity_df(entity_schema)
        )


        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df,
            entity_df_event_timestamp_col,
            config,
        )

        @contextlib.contextmanager
        def query_generator() -> Iterator[str]:
            table_name = offline_utils.get_temp_entity_table_name()

            _upload_entity_df(config, entity_df, table_name)

            expected_join_keys = offline_utils.get_expected_join_keys(
                project, feature_views, registry
            )

            offline_utils.assert_expected_columns_in_entity_df(
                entity_schema, expected_join_keys, entity_df_event_timestamp_col
            )

            query_context = offline_utils.get_feature_view_query_context(
                feature_refs,
                feature_views,
                registry,
                project,
                entity_df_event_timestamp_range,
            )

            query_context_dict = [asdict(context) for context in query_context]
            for context in query_context_dict:
                context["entity_selections"] = [
                    f"""{entity_selection.replace(' AS ', ' AS ')}"""
                    for entity_selection in context["entity_selections"]
                ]

            try:
                yield build_point_in_time_query(
                    query_context_dict,
                    left_table_query_string=table_name,
                    entity_df_event_timestamp_col=entity_df_event_timestamp_col,
                    entity_df_columns=entity_schema.keys(),
                    query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                    full_feature_names=full_feature_names,
                )
            finally:
                if table_name:
                    with _get_conn(config.offline_store) as conn, conn.cursor() as cur:
                        cur.execute(
                            f"""
                                DROP TABLE IF EXISTS {table_name};
                                """
                        )

        return DorisSQLRetrievalJob(
            query=query_generator,
            config=config,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(entity_schema.keys() - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="doris")
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, DorisSQLOfflineStoreConfig)
        assert isinstance(data_source, DorisSQLSource)
        from_expression = data_source.get_table_query_string()

        field_string = ", ".join(
            join_key_columns + feature_name_columns + [timestamp_field]
        )

        start_date = start_date.astimezone(tz=utc)
        end_date = end_date.astimezone(tz=utc)

        query = f"""
            SELECT {field_string}
            FROM {from_expression} AS paftoq_alias
            WHERE {timestamp_field} BETWEEN '{start_date}' AND '{end_date}'
        """

        return DorisSQLRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
        )


class DorisSQLRetrievalJob(RetrievalJob):
    def __init__(
        self,
        query: Union[str, Callable[[], ContextManager[str]]],
        config: RepoConfig,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        if not isinstance(query, str):
            self._query_generator = query
        else:

            @contextlib.contextmanager
            def query_generator() -> Iterator[str]:
                assert isinstance(query, str)
                yield query

            self._query_generator = query_generator
        self.config = config
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        return self._to_arrow_internal().to_pandas()

    def to_sql(self) -> str:
        with self._query_generator() as query:
            return query

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        with self._query_generator() as query:
            with _get_conn(self.config.offline_store) as conn, conn.cursor() as cur:
                # print(query)
                cur.execute(query)
                fields = [
                    (c[0], doris_type_code_to_arrow(c[1])) for c in cur.description
                ]
                data = cur.fetchall()
                schema = pa.schema(fields)
                # TODO: Fix...
                data_transposed: List[List[Any]] = []
                for col in range(len(fields)):
                    data_transposed.append([])
                    for row in range(len(data)):
                        data_transposed[col].append(data[row][col])

                table = pa.Table.from_arrays(
                    [pa.array(row) for row in data_transposed], schema=schema
                )
                return table

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pd.DataFrame, str],
    entity_df_event_timestamp_col: str,
    config: RepoConfig,
) -> Tuple[datetime, datetime]:
    if isinstance(entity_df, pd.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pd.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pd.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    elif isinstance(entity_df, str):
        with _get_conn(config.offline_store) as conn, conn.cursor() as cur:
            cur.execute(f"SELECT MIN({entity_df_event_timestamp_col}) AS min, MAX({entity_df_event_timestamp_col}) AS max FROM ({entity_df}) as tmp_alias")
            res = cur.fetchall()
        entity_df_event_timestamp_range = (
            pd.to_datetime(res[0][0]),
            pd.to_datetime(res[0][1]),
        )
    else:
        raise InvalidEntityType(type(entity_df))

    return entity_df_event_timestamp_range


def _append_alias(field_names: List[str], alias: str) -> List[str]:
    return [f"{alias}.{field_name}" for field_name in field_names]


def build_point_in_time_query(
    feature_view_query_contexts: List[dict],
    left_table_query_string: str,
    entity_df_event_timestamp_col: str,
    entity_df_columns: KeysView[str],
    query_template: str,
    full_feature_names: bool = False,
) -> str:
    """Build point-in-time query between each feature view table and the entity dataframe for PostgreSQL"""
    template = Environment(loader=BaseLoader()).from_string(source=query_template)

    final_output_feature_names = list(entity_df_columns)
    final_output_feature_names.extend(
        [
            (
                f'{fv["name"]}__{fv["field_mapping"].get(feature, feature)}'
                if full_feature_names
                else fv["field_mapping"].get(feature, feature)
            )
            for fv in feature_view_query_contexts
            for feature in fv["features"]
        ]
    )

    # Add additional fields to dict
    template_context = {
        "left_table_query_string": left_table_query_string,
        "entity_df_event_timestamp_col": entity_df_event_timestamp_col,
        "unique_entity_keys": set(
            [entity for fv in feature_view_query_contexts for entity in fv["entities"]]
        ),
        "featureviews": feature_view_query_contexts,
        "full_feature_names": full_feature_names,
        "final_output_feature_names": final_output_feature_names,
    }

    query = template.render(template_context)
    return query


def _upload_entity_df(
    config: RepoConfig, entity_df: Union[pd.DataFrame, str], table_name: str
):
    if isinstance(entity_df, pd.DataFrame):
        df_to_doris_table(config.offline_store, entity_df, table_name)
    elif isinstance(entity_df, str):
        with _get_conn(config.offline_store) as conn, conn.cursor() as cursor:
            cursor.execute(f"({entity_df})")
            columns = cursor.description
            cursor.fetchall()
            type_mapping = {
                1: "BOOLEAN",
                2: "SMALLINT",
                3: "INT",
                8: "BIGINT",
                254: "VARCHAR(65533)",
                4: "FLOAT",
                5: "DOUBLE",
                10: "DATE",
                12: "DATETIME",
                253: "VARCHAR(65533)",
                252: "VARCHAR(65533)",
                246: "DECIMAL(38,37)",
                245: "DECIMAL(38,37)",
            }
            column_definitions = []
            for column in columns:
                column_name = column[0]
                column_type = column[1]
                sql_type = type_mapping.get(column_type, "STRING")
                column_definitions.append(f"{column_name} {sql_type}")
                columns_definition_str = ",\n  ".join(column_definitions)

            create_table_query = f"""
                CREATE TABLE {table_name} (
                {columns_definition_str}
                )
                DISTRIBUTED BY HASH(event_timestamp) BUCKETS 1
                PROPERTIES (
                'replication_num' = '1'
                );
            """
            # print(create_table_query)
            cursor.execute(create_table_query)
            insert_data_query = f"""
            INSERT INTO {table_name}
            SELECT *
            FROM ({entity_df}) AS sub;
            """
            cursor.execute(insert_data_query)
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_schema(
    entity_df: Union[pd.DataFrame, str],
    config: RepoConfig,
) -> Dict[str, np.dtype]:
    if isinstance(entity_df, pd.DataFrame):
        return dict(zip(entity_df.columns, entity_df.dtypes))

    elif isinstance(entity_df, str):
        df_query = f"({entity_df})"
        return get_query_schema(config.offline_store, df_query)
    else:
        raise InvalidEntityType(type(entity_df))


MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
/*
 Compute a deterministic hash for the `left_table_query_string` that will be used throughout
 all the logic as the field to GROUP BY the data
*/
WITH entity_dataframe AS (
    SELECT *,
        {{entity_df_event_timestamp_col}} AS entity_timestamp
        {% for featureview in featureviews %}
            ,CONCAT(
                {% for entity_key in unique_entity_keys %}
                    {{entity_key}},
                {% endfor %}
                {{entity_df_event_timestamp_col}}
            ) AS {{featureview.name}}__entity_row_unique_id
        {% endfor %}
    FROM {{ left_table_query_string }}
),

{% for featureview in featureviews %}

{{ featureview.name }}__entity_dataframe AS (
    SELECT
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
    FROM entity_dataframe
    GROUP BY
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
),

/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.

 1. We first join the current feature_view to the entity dataframe that has been passed.
 This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously

 The output of this CTE will contain all the necessary information and already filtered out most
 of the data that is not relevant.
*/

{{ featureview.name }}__subquery AS (
    SELECT
        {{ featureview.timestamp_field }} as event_timestamp,
        {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
        {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% endif %}
        {% for feature in featureview.features %}
            {{ feature }} as {% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}{% if loop.last %}{% else %}, {% endif %}
        {% endfor %}
    FROM {{ featureview.table_subquery }} AS sub
    WHERE {{ featureview.timestamp_field }} <= '{{ featureview.max_event_timestamp }}'
    {% if featureview.ttl != 0 %}
    AND {{ featureview.timestamp_field }} >= '{{ featureview.min_event_timestamp }}'
    {% endif %}
),

{{ featureview.name }}__base AS (
    SELECT
        subquery.*,
        entity_dataframe.{{entity_df_event_timestamp_col}} AS entity_timestamp,
        entity_dataframe.{{featureview.name}}__entity_row_unique_id
    FROM {{ featureview.name }}__subquery AS subquery
    INNER JOIN entity_dataframe
        ON 1=1
        AND subquery.event_timestamp <= entity_dataframe.{{entity_df_event_timestamp_col}}
        {% if featureview.ttl != 0 %}
        AND {{ featureview.ttl }} >= TIMESTAMPDIFF(SECOND, subquery.event_timestamp, entity_dataframe.{{entity_df_event_timestamp_col}})
        {% endif %}
        {% for entity in featureview.entities %}
        AND subquery.{{ entity }} = entity_dataframe.{{ entity }}
        {% endfor %}
),

/*
 2. If the `created_timestamp_column` has been set, we need to
 deduplicate the data first. This is done by calculating the
 `MAX(created_at_timestamp)` for each event_timestamp.
 We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
{{ featureview.name }}__dedup AS (
    SELECT
        {{featureview.name}}__entity_row_unique_id,
        event_timestamp,
        MAX(created_timestamp) as created_timestamp
    FROM {{ featureview.name }}__base
    GROUP BY {{featureview.name}}__entity_row_unique_id, event_timestamp
),
{% endif %}

/*
 3. The data has been filtered during the first CTE "*__base"
 Thus we only need to compute the latest timestamp of each feature.
*/
{{ featureview.name }}__latest AS (
    SELECT
        {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id,
        MAX({{ featureview.name }}__base.event_timestamp) AS event_timestamp
        {% if featureview.created_timestamp_column %}
            ,MAX({{ featureview.name }}__base.created_timestamp) AS created_timestamp
        {% endif %}
    FROM {{ featureview.name }}__base
    {% if featureview.created_timestamp_column %}
        INNER JOIN {{ featureview.name }}__dedup
        ON {{ featureview.name }}__dedup.{{ featureview.name }}__entity_row_unique_id = {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id
        AND {{ featureview.name }}__dedup.event_timestamp = {{ featureview.name }}__base.event_timestamp
        AND {{ featureview.name }}__dedup.created_timestamp = {{ featureview.name }}__base.created_timestamp
    {% endif %}
    GROUP BY {{ featureview.name }}__base.{{ featureview.name }}__entity_row_unique_id
),

/*
 4. Once we know the latest value of each feature for a given timestamp,
 we can join again the data back to the original "base" dataset
*/
{{ featureview.name }}__cleaned AS (
    SELECT base.*
    FROM {{ featureview.name }}__base as base
    INNER JOIN {{ featureview.name }}__latest
    ON base.{{ featureview.name }}__entity_row_unique_id = {{ featureview.name }}__latest.{{ featureview.name }}__entity_row_unique_id
    AND base.event_timestamp = {{ featureview.name }}__latest.event_timestamp
        {% if featureview.created_timestamp_column %}
            AND base.created_timestamp = {{ featureview.name }}__latest.created_timestamp
        {% endif %}
){% if loop.last %}{% else %}, {% endif %}

{% endfor %}

/*
 Joins the outputs of multiple time travel joins to a single table.
 The entity_dataframe dataset being our source of truth here.
 */

SELECT {{ final_output_feature_names | join(', ')}}
FROM entity_dataframe
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
        {{featureview.name}}__entity_row_unique_id
        {% for feature in featureview.features %}
            ,{% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}
        {% endfor %}
    FROM {{ featureview.name }}__cleaned
) AS {{ featureview.name }}_cleaned
ON
{{ featureview.name }}_cleaned.{{ featureview.name }}__entity_row_unique_id = entity_dataframe.{{ featureview.name }}__entity_row_unique_id
{% endfor %}
"""



import logging
import os
import uuid
from binascii import hexlify
from datetime import datetime, timedelta
from enum import Enum
from threading import Lock
from typing import Any, Callable, List, Literal, Optional, Set, Union

from pydantic import ConfigDict, Field, StrictStr

import feast
from feast import usage
from feast.base_feature_view import BaseFeatureView
from feast.data_source import DataSource
from feast.entity import Entity
from feast.errors import (
    DataSourceObjectNotFoundException,
    EntityNotFoundException,
    FeatureServiceNotFoundException,
    FeatureViewNotFoundException,
    SavedDatasetNotFound,
    ValidationReferenceNotFound,
)
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.infra.infra_object import Infra
from feast.infra.registry import proto_registry_utils
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.utils.snowflake.snowflake_utils import (
    GetSnowflakeConnection,
    execute_snowflake_statement,
)
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.project_metadata import ProjectMetadata
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.Entity_pb2 import Entity as EntityProto
from feast.protos.feast.core.FeatureService_pb2 import (
    FeatureService as FeatureServiceProto,
)
from feast.protos.feast.core.FeatureView_pb2 import FeatureView as FeatureViewProto
from feast.protos.feast.core.InfraObject_pb2 import Infra as InfraProto
from feast.protos.feast.core.OnDemandFeatureView_pb2 import (
    OnDemandFeatureView as OnDemandFeatureViewProto,
)
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.protos.feast.core.SavedDataset_pb2 import SavedDataset as SavedDatasetProto
from feast.protos.feast.core.StreamFeatureView_pb2 import (
    StreamFeatureView as StreamFeatureViewProto,
)
from feast.protos.feast.core.ValidationProfile_pb2 import (
    ValidationReference as ValidationReferenceProto,
)
from feast.repo_config import RegistryConfig
from feast.saved_dataset import SavedDataset, ValidationReference
from feast.stream_feature_view import StreamFeatureView

logger = logging.getLogger(__name__)


class FeastMetadataKeys(Enum):
    LAST_UPDATED_TIMESTAMP = "last_updated_timestamp"
    PROJECT_UUID = "project_uuid"


class SnowflakeRegistryConfig(RegistryConfig):
    """Registry config for Snowflake"""

    registry_type: Literal["snowflake.registry"] = "snowflake.registry"
    """ Registry type selector """

    type: Literal["snowflake.registry"] = "snowflake.registry"
    """ Registry type selector """

    config_path: Optional[str] = os.path.expanduser("~/.snowsql/config")
    """ Snowflake config path -- absolute path required (Cant use ~) """

    account: Optional[str] = None
    """ Snowflake deployment identifier -- drop .snowflakecomputing.com """

    user: Optional[str] = None
    """ Snowflake user name """

    password: Optional[str] = None
    """ Snowflake password """

    role: Optional[str] = None
    """ Snowflake role name """

    warehouse: Optional[str] = None
    """ Snowflake warehouse name """

    authenticator: Optional[str] = None
    """ Snowflake authenticator name """

    database: StrictStr
    """ Snowflake database name """

    schema_: Optional[str] = Field("PUBLIC", alias="schema")
    """ Snowflake schema name """
    model_config = ConfigDict(populate_by_name=True)


class SnowflakeRegistry(BaseRegistry):
    def __init__(
        self,
        registry_config,
        project: str,
        repo_path,
    ):
        assert registry_config is not None and isinstance(
            registry_config, SnowflakeRegistryConfig
        ), "SnowflakeRegistry needs a valid registry_config, a path does not work"

        self.registry_config = registry_config
        self.registry_path = (
            f'"{self.registry_config.database}"."{self.registry_config.schema_}"'
        )

        with GetSnowflakeConnection(self.registry_config) as conn:
            sql_function_file = f"{os.path.dirname(feast.__file__)}/infra/utils/snowflake/registry/snowflake_table_creation.sql"
            with open(sql_function_file, "r") as file:
                sqlFile = file.read()

                sqlCommands = sqlFile.split(";")
                for command in sqlCommands:
                    query = command.replace("REGISTRY_PATH", f"{self.registry_path}")
                    execute_snowflake_statement(conn, query)

        self.cached_registry_proto = self.proto()
        proto_registry_utils.init_project_metadata(self.cached_registry_proto, project)
        self.cached_registry_proto_created = datetime.utcnow()
        self._refresh_lock = Lock()
        self.cached_registry_proto_ttl = timedelta(
            seconds=registry_config.cache_ttl_seconds
            if registry_config.cache_ttl_seconds is not None
            else 0
        )
        self.project = project

    def refresh(self, project: Optional[str] = None):
        if project:
            project_metadata = proto_registry_utils.get_project_metadata(
                registry_proto=self.cached_registry_proto, project=project
            )
            if project_metadata:
                usage.set_current_project_uuid(project_metadata.project_uuid)
            else:
                proto_registry_utils.init_project_metadata(
                    self.cached_registry_proto, project
                )
        self.cached_registry_proto = self.proto()
        self.cached_registry_proto_created = datetime.utcnow()

    def _refresh_cached_registry_if_necessary(self):
        with self._refresh_lock:
            expired = (
                self.cached_registry_proto is None
                or self.cached_registry_proto_created is None
            ) or (
                self.cached_registry_proto_ttl.total_seconds()
                > 0  # 0 ttl means infinity
                and (
                    datetime.utcnow()
                    > (
                        self.cached_registry_proto_created
                        + self.cached_registry_proto_ttl
                    )
                )
            )

            if expired:
                logger.info("Registry cache expired, so refreshing")
                self.refresh()

    def teardown(self):
        with GetSnowflakeConnection(self.registry_config) as conn:
            sql_function_file = f"{os.path.dirname(feast.__file__)}/infra/utils/snowflake/registry/snowflake_table_deletion.sql"
            with open(sql_function_file, "r") as file:
                sqlFile = file.read()

                sqlCommands = sqlFile.split(";")
                for command in sqlCommands:
                    query = command.replace("REGISTRY_PATH", f"{self.registry_path}")
                    execute_snowflake_statement(conn, query)

    # apply operations
    def apply_data_source(
        self, data_source: DataSource, project: str, commit: bool = True
    ):
        return self._apply_object(
            "DATA_SOURCES",
            project,
            "DATA_SOURCE_NAME",
            data_source,
            "DATA_SOURCE_PROTO",
        )

    def apply_entity(self, entity: Entity, project: str, commit: bool = True):
        return self._apply_object(
            "ENTITIES", project, "ENTITY_NAME", entity, "ENTITY_PROTO"
        )

    def apply_feature_service(
        self, feature_service: FeatureService, project: str, commit: bool = True
    ):
        return self._apply_object(
            "FEATURE_SERVICES",
            project,
            "FEATURE_SERVICE_NAME",
            feature_service,
            "FEATURE_SERVICE_PROTO",
        )

    def apply_feature_view(
        self, feature_view: BaseFeatureView, project: str, commit: bool = True
    ):
        fv_table_str = self._infer_fv_table(feature_view)
        fv_column_name = fv_table_str[:-1]
        return self._apply_object(
            fv_table_str,
            project,
            f"{fv_column_name}_NAME",
            feature_view,
            f"{fv_column_name}_PROTO",
        )

    def apply_saved_dataset(
        self,
        saved_dataset: SavedDataset,
        project: str,
        commit: bool = True,
    ):
        return self._apply_object(
            "SAVED_DATASETS",
            project,
            "SAVED_DATASET_NAME",
            saved_dataset,
            "SAVED_DATASET_PROTO",
        )

    def apply_validation_reference(
        self,
        validation_reference: ValidationReference,
        project: str,
        commit: bool = True,
    ):
        return self._apply_object(
            "VALIDATION_REFERENCES",
            project,
            "VALIDATION_REFERENCE_NAME",
            validation_reference,
            "VALIDATION_REFERENCE_PROTO",
        )

    def update_infra(self, infra: Infra, project: str, commit: bool = True):
        self._apply_object(
            "MANAGED_INFRA",
            project,
            "INFRA_NAME",
            infra,
            "INFRA_PROTO",
            name="infra_obj",
        )

    def _apply_object(
        self,
        table: str,
        project: str,
        id_field_name: str,
        obj: Any,
        proto_field_name: str,
        name: Optional[str] = None,
    ):
        self._maybe_init_project_metadata(project)

        name = name or (obj.name if hasattr(obj, "name") else None)
        assert name, f"name needs to be provided for {obj}"

        update_datetime = datetime.utcnow()
        if hasattr(obj, "last_updated_timestamp"):
            obj.last_updated_timestamp = update_datetime

        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    project_id
                FROM
                    {self.registry_path}."{table}"
                WHERE
                    project_id = '{project}'
                    AND {id_field_name.lower()} = '{name}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            if not df.empty:
                proto = hexlify(obj.to_proto().SerializeToString()).__str__()[1:]
                query = f"""
                    UPDATE {self.registry_path}."{table}"
                        SET
                            {proto_field_name} = TO_BINARY({proto}),
                            last_updated_timestamp = CURRENT_TIMESTAMP()
                        WHERE
                            {id_field_name.lower()} = '{name}'
                """
                execute_snowflake_statement(conn, query)

            else:
                obj_proto = obj.to_proto()

                if hasattr(obj_proto, "meta") and hasattr(
                    obj_proto.meta, "created_timestamp"
                ):
                    obj_proto.meta.created_timestamp.FromDatetime(update_datetime)

                proto = hexlify(obj_proto.SerializeToString()).__str__()[1:]
                if table == "FEATURE_VIEWS":
                    query = f"""
                        INSERT INTO {self.registry_path}."{table}"
                            VALUES
                            ('{name}', '{project}', CURRENT_TIMESTAMP(), TO_BINARY({proto}), '', '')
                    """
                elif "_FEATURE_VIEWS" in table:
                    query = f"""
                        INSERT INTO {self.registry_path}."{table}"
                            VALUES
                            ('{name}', '{project}', CURRENT_TIMESTAMP(), TO_BINARY({proto}), '')
                    """
                else:
                    query = f"""
                        INSERT INTO {self.registry_path}."{table}"
                            VALUES
                            ('{name}', '{project}', CURRENT_TIMESTAMP(), TO_BINARY({proto}))
                    """
                execute_snowflake_statement(conn, query)

            self._set_last_updated_metadata(update_datetime, project)

    # delete operations
    def delete_data_source(self, name: str, project: str, commit: bool = True):
        return self._delete_object(
            "DATA_SOURCES",
            name,
            project,
            "DATA_SOURCE_NAME",
            DataSourceObjectNotFoundException,
        )

    def delete_entity(self, name: str, project: str, commit: bool = True):
        return self._delete_object(
            "ENTITIES", name, project, "ENTITY_NAME", EntityNotFoundException
        )

    def delete_feature_service(self, name: str, project: str, commit: bool = True):
        return self._delete_object(
            "FEATURE_SERVICES",
            name,
            project,
            "FEATURE_SERVICE_NAME",
            FeatureServiceNotFoundException,
        )

    # can you have featureviews with the same name
    def delete_feature_view(self, name: str, project: str, commit: bool = True):
        deleted_count = 0
        for table in {
            "FEATURE_VIEWS",
            "ON_DEMAND_FEATURE_VIEWS",
            "STREAM_FEATURE_VIEWS",
        }:
            deleted_count += self._delete_object(
                table, name, project, "FEATURE_VIEW_NAME", None
            )
        if deleted_count == 0:
            raise FeatureViewNotFoundException(name, project)

    def delete_saved_dataset(self, name: str, project: str, allow_cache: bool = False):
        self._delete_object(
            "SAVED_DATASETS",
            name,
            project,
            "SAVED_DATASET_NAME",
            SavedDatasetNotFound,
        )

    def delete_validation_reference(self, name: str, project: str, commit: bool = True):
        self._delete_object(
            "VALIDATION_REFERENCES",
            name,
            project,
            "VALIDATION_REFERENCE_NAME",
            ValidationReferenceNotFound,
        )

    def _delete_object(
        self,
        table: str,
        name: str,
        project: str,
        id_field_name: str,
        not_found_exception: Optional[Callable],
    ):
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                DELETE FROM {self.registry_path}."{table}"
                WHERE
                    project_id = '{project}'
                    AND {id_field_name.lower()} = '{name}'
            """
            cursor = execute_snowflake_statement(conn, query)

            if cursor.rowcount < 1 and not_found_exception:  # type: ignore
                raise not_found_exception(name, project)
            self._set_last_updated_metadata(datetime.utcnow(), project)

            return cursor.rowcount

    # get operations
    def get_data_source(
        self, name: str, project: str, allow_cache: bool = False
    ) -> DataSource:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_data_source(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "DATA_SOURCES",
            name,
            project,
            DataSourceProto,
            DataSource,
            "DATA_SOURCE_NAME",
            "DATA_SOURCE_PROTO",
            DataSourceObjectNotFoundException,
        )

    def get_entity(self, name: str, project: str, allow_cache: bool = False) -> Entity:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_entity(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "ENTITIES",
            name,
            project,
            EntityProto,
            Entity,
            "ENTITY_NAME",
            "ENTITY_PROTO",
            EntityNotFoundException,
        )

    def get_feature_service(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureService:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_feature_service(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "FEATURE_SERVICES",
            name,
            project,
            FeatureServiceProto,
            FeatureService,
            "FEATURE_SERVICE_NAME",
            "FEATURE_SERVICE_PROTO",
            FeatureServiceNotFoundException,
        )

    def get_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureView:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "FEATURE_VIEWS",
            name,
            project,
            FeatureViewProto,
            FeatureView,
            "FEATURE_VIEW_NAME",
            "FEATURE_VIEW_PROTO",
            FeatureViewNotFoundException,
        )

    def get_infra(self, project: str, allow_cache: bool = False) -> Infra:
        infra_object = self._get_object(
            "MANAGED_INFRA",
            "infra_obj",
            project,
            InfraProto,
            Infra,
            "INFRA_NAME",
            "INFRA_PROTO",
            None,
        )
        infra_object = infra_object or InfraProto()
        return Infra.from_proto(infra_object)

    def get_on_demand_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> OnDemandFeatureView:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_on_demand_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "ON_DEMAND_FEATURE_VIEWS",
            name,
            project,
            OnDemandFeatureViewProto,
            OnDemandFeatureView,
            "ON_DEMAND_FEATURE_VIEW_NAME",
            "ON_DEMAND_FEATURE_VIEW_PROTO",
            FeatureViewNotFoundException,
        )

    def get_saved_dataset(
        self, name: str, project: str, allow_cache: bool = False
    ) -> SavedDataset:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_saved_dataset(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "SAVED_DATASETS",
            name,
            project,
            SavedDatasetProto,
            SavedDataset,
            "SAVED_DATASET_NAME",
            "SAVED_DATASET_PROTO",
            SavedDatasetNotFound,
        )

    def get_stream_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ):
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_stream_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "STREAM_FEATURE_VIEWS",
            name,
            project,
            StreamFeatureViewProto,
            StreamFeatureView,
            "STREAM_FEATURE_VIEW_NAME",
            "STREAM_FEATURE_VIEW_PROTO",
            FeatureViewNotFoundException,
        )

    def get_validation_reference(
        self, name: str, project: str, allow_cache: bool = False
    ) -> ValidationReference:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_validation_reference(
                self.cached_registry_proto, name, project
            )
        return self._get_object(
            "VALIDATION_REFERENCES",
            name,
            project,
            ValidationReferenceProto,
            ValidationReference,
            "VALIDATION_REFERENCE_NAME",
            "VALIDATION_REFERENCE_PROTO",
            ValidationReferenceNotFound,
        )

    def _get_object(
        self,
        table: str,
        name: str,
        project: str,
        proto_class: Any,
        python_class: Any,
        id_field_name: str,
        proto_field_name: str,
        not_found_exception: Optional[Callable],
    ):
        self._maybe_init_project_metadata(project)
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    {proto_field_name}
                FROM
                    {self.registry_path}."{table}"
                WHERE
                    project_id = '{project}'
                    AND {id_field_name.lower()} = '{name}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

        if not df.empty:
            _proto = proto_class.FromString(df.squeeze())
            return python_class.from_proto(_proto)
        elif not_found_exception:
            raise not_found_exception(name, project)
        else:
            return None

    # list operations
    def list_data_sources(
        self, project: str, allow_cache: bool = False
    ) -> List[DataSource]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_data_sources(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "DATA_SOURCES", project, DataSourceProto, DataSource, "DATA_SOURCE_PROTO"
        )

    def list_entities(self, project: str, allow_cache: bool = False) -> List[Entity]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_entities(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "ENTITIES", project, EntityProto, Entity, "ENTITY_PROTO"
        )

    def list_feature_services(
        self, project: str, allow_cache: bool = False
    ) -> List[FeatureService]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_feature_services(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "FEATURE_SERVICES",
            project,
            FeatureServiceProto,
            FeatureService,
            "FEATURE_SERVICE_PROTO",
        )

    def list_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[FeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "FEATURE_VIEWS",
            project,
            FeatureViewProto,
            FeatureView,
            "FEATURE_VIEW_PROTO",
        )

    def list_on_demand_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[OnDemandFeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_on_demand_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "ON_DEMAND_FEATURE_VIEWS",
            project,
            OnDemandFeatureViewProto,
            OnDemandFeatureView,
            "ON_DEMAND_FEATURE_VIEW_PROTO",
        )

    def list_saved_datasets(
        self, project: str, allow_cache: bool = False
    ) -> List[SavedDataset]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_saved_datasets(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "SAVED_DATASETS",
            project,
            SavedDatasetProto,
            SavedDataset,
            "SAVED_DATASET_PROTO",
        )

    def list_stream_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[StreamFeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_stream_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_objects(
            "STREAM_FEATURE_VIEWS",
            project,
            StreamFeatureViewProto,
            StreamFeatureView,
            "STREAM_FEATURE_VIEW_PROTO",
        )

    def list_validation_references(
        self, project: str, allow_cache: bool = False
    ) -> List[ValidationReference]:
        return self._list_objects(
            "VALIDATION_REFERENCES",
            project,
            ValidationReferenceProto,
            ValidationReference,
            "VALIDATION_REFERENCE_PROTO",
        )

    def _list_objects(
        self,
        table: str,
        project: str,
        proto_class: Any,
        python_class: Any,
        proto_field_name: str,
    ):
        self._maybe_init_project_metadata(project)
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    {proto_field_name}
                FROM
                    {self.registry_path}."{table}"
                WHERE
                    project_id = '{project}'
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            if not df.empty:
                return [
                    python_class.from_proto(
                        proto_class.FromString(row[1][proto_field_name])
                    )
                    for row in df.iterrows()
                ]
        return []

    def apply_materialization(
        self,
        feature_view: FeatureView,
        project: str,
        start_date: datetime,
        end_date: datetime,
        commit: bool = True,
    ):
        fv_table_str = self._infer_fv_table(feature_view)
        fv_column_name = fv_table_str[:-1]
        python_class, proto_class = self._infer_fv_classes(feature_view)

        if python_class in {OnDemandFeatureView}:
            raise ValueError(
                f"Cannot apply materialization for feature {feature_view.name} of type {python_class}"
            )
        fv: Union[FeatureView, StreamFeatureView] = self._get_object(
            fv_table_str,
            feature_view.name,
            project,
            proto_class,
            python_class,
            f"{fv_column_name}_NAME",
            f"{fv_column_name}_PROTO",
            FeatureViewNotFoundException,
        )
        fv.materialization_intervals.append((start_date, end_date))
        self._apply_object(
            fv_table_str,
            project,
            f"{fv_column_name}_NAME",
            fv,
            f"{fv_column_name}_PROTO",
        )

    def list_project_metadata(
        self, project: str, allow_cache: bool = False
    ) -> List[ProjectMetadata]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_project_metadata(
                self.cached_registry_proto, project
            )
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    metadata_key,
                    metadata_value
                FROM
                    {self.registry_path}."FEAST_METADATA"
                WHERE
                    project_id = '{project}'
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            if not df.empty:
                project_metadata = ProjectMetadata(project_name=project)
                for row in df.iterrows():
                    if row[1]["METADATA_KEY"] == FeastMetadataKeys.PROJECT_UUID.value:
                        project_metadata.project_uuid = row[1]["METADATA_VALUE"]
                        break
                    # TODO(adchia): Add other project metadata in a structured way
                return [project_metadata]
        return []

    def apply_user_metadata(
        self,
        project: str,
        feature_view: BaseFeatureView,
        metadata_bytes: Optional[bytes],
    ):
        fv_table_str = self._infer_fv_table(feature_view)
        fv_column_name = fv_table_str[:-1].lower()
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    project_id
                FROM
                    {self.registry_path}."{fv_table_str}"
                WHERE
                    project_id = '{project}'
                    AND {fv_column_name}_name = '{feature_view.name}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            if not df.empty:
                if metadata_bytes:
                    metadata_hex = hexlify(metadata_bytes).__str__()[1:]
                else:
                    metadata_hex = "''"
                query = f"""
                    UPDATE {self.registry_path}."{fv_table_str}"
                        SET
                            user_metadata = TO_BINARY({metadata_hex}),
                            last_updated_timestamp = CURRENT_TIMESTAMP()
                        WHERE
                            project_id = '{project}'
                            AND {fv_column_name}_name = '{feature_view.name}'
                """
                execute_snowflake_statement(conn, query)
            else:
                raise FeatureViewNotFoundException(feature_view.name, project=project)

    def get_user_metadata(
        self, project: str, feature_view: BaseFeatureView
    ) -> Optional[bytes]:
        fv_table_str = self._infer_fv_table(feature_view)
        fv_column_name = fv_table_str[:-1].lower()
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    user_metadata
                FROM
                    {self.registry_path}."{fv_table_str}"
                WHERE
                    {fv_column_name}_name = '{feature_view.name}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

        if not df.empty:
            return df.squeeze()
        else:
            raise FeatureViewNotFoundException(feature_view.name, project=project)

    def proto(self) -> RegistryProto:
        r = RegistryProto()
        last_updated_timestamps = []
        projects = self._get_all_projects()
        for project in projects:
            for lister, registry_proto_field in [
                (self.list_entities, r.entities),
                (self.list_feature_views, r.feature_views),
                (self.list_data_sources, r.data_sources),
                (self.list_on_demand_feature_views, r.on_demand_feature_views),
                (self.list_stream_feature_views, r.stream_feature_views),
                (self.list_feature_services, r.feature_services),
                (self.list_saved_datasets, r.saved_datasets),
                (self.list_validation_references, r.validation_references),
                (self.list_project_metadata, r.project_metadata),
            ]:
                objs: List[Any] = lister(project)  # type: ignore
                if objs:
                    obj_protos = [obj.to_proto() for obj in objs]
                    for obj_proto in obj_protos:
                        if "spec" in obj_proto.DESCRIPTOR.fields_by_name:
                            obj_proto.spec.project = project
                        else:
                            obj_proto.project = project
                    registry_proto_field.extend(obj_protos)

            # This is suuuper jank. Because of https://github.com/feast-dev/feast/issues/2783,
            # the registry proto only has a single infra field, which we're currently setting as the "last" project.
            r.infra.CopyFrom(self.get_infra(project).to_proto())
            last_updated_timestamps.append(self._get_last_updated_metadata(project))

        if last_updated_timestamps:
            r.last_updated.FromDatetime(max(last_updated_timestamps))

        return r

    def _get_all_projects(self) -> Set[str]:
        projects = set()

        base_tables = [
            "DATA_SOURCES",
            "ENTITIES",
            "FEATURE_VIEWS",
            "ON_DEMAND_FEATURE_VIEWS",
            "STREAM_FEATURE_VIEWS",
        ]

        with GetSnowflakeConnection(self.registry_config) as conn:
            for table in base_tables:
                query = (
                    f'SELECT DISTINCT project_id FROM {self.registry_path}."{table}"'
                )
                df = execute_snowflake_statement(conn, query).fetch_pandas_all()

                for row in df.iterrows():
                    projects.add(row[1]["PROJECT_ID"])

        return projects

    def _get_last_updated_metadata(self, project: str):
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    metadata_value
                FROM
                    {self.registry_path}."FEAST_METADATA"
                WHERE
                    project_id = '{project}'
                    AND metadata_key = '{FeastMetadataKeys.LAST_UPDATED_TIMESTAMP.value}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

        if df.empty:
            return None

        return datetime.utcfromtimestamp(int(df.squeeze()))

    def _infer_fv_classes(self, feature_view):
        if isinstance(feature_view, StreamFeatureView):
            python_class, proto_class = StreamFeatureView, StreamFeatureViewProto
        elif isinstance(feature_view, FeatureView):
            python_class, proto_class = FeatureView, FeatureViewProto
        elif isinstance(feature_view, OnDemandFeatureView):
            python_class, proto_class = OnDemandFeatureView, OnDemandFeatureViewProto
        else:
            raise ValueError(f"Unexpected feature view type: {type(feature_view)}")
        return python_class, proto_class

    def _infer_fv_table(self, feature_view) -> str:
        if isinstance(feature_view, StreamFeatureView):
            table = "STREAM_FEATURE_VIEWS"
        elif isinstance(feature_view, FeatureView):
            table = "FEATURE_VIEWS"
        elif isinstance(feature_view, OnDemandFeatureView):
            table = "ON_DEMAND_FEATURE_VIEWS"
        else:
            raise ValueError(f"Unexpected feature view type: {type(feature_view)}")
        return table

    def _maybe_init_project_metadata(self, project):
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    metadata_value
                FROM
                    {self.registry_path}."FEAST_METADATA"
                WHERE
                    project_id = '{project}'
                    AND metadata_key = '{FeastMetadataKeys.PROJECT_UUID.value}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            if not df.empty:
                usage.set_current_project_uuid(df.squeeze())
            else:
                new_project_uuid = f"{uuid.uuid4()}"
                query = f"""
                    INSERT INTO {self.registry_path}."FEAST_METADATA"
                        VALUES
                        ('{project}', '{FeastMetadataKeys.PROJECT_UUID.value}', '{new_project_uuid}', CURRENT_TIMESTAMP())
                """
                execute_snowflake_statement(conn, query)

                usage.set_current_project_uuid(new_project_uuid)

    def _set_last_updated_metadata(self, last_updated: datetime, project: str):
        with GetSnowflakeConnection(self.registry_config) as conn:
            query = f"""
                SELECT
                    project_id
                FROM
                    {self.registry_path}."FEAST_METADATA"
                WHERE
                    project_id = '{project}'
                    AND metadata_key = '{FeastMetadataKeys.LAST_UPDATED_TIMESTAMP.value}'
                LIMIT 1
            """
            df = execute_snowflake_statement(conn, query).fetch_pandas_all()

            update_time = int(last_updated.timestamp())
            if not df.empty:
                query = f"""
                    UPDATE {self.registry_path}."FEAST_METADATA"
                        SET
                            project_id = '{project}',
                            metadata_key = '{FeastMetadataKeys.LAST_UPDATED_TIMESTAMP.value}',
                            metadata_value = '{update_time}',
                            last_updated_timestamp = CURRENT_TIMESTAMP()
                        WHERE
                            project_id = '{project}'
                            AND metadata_key = '{FeastMetadataKeys.LAST_UPDATED_TIMESTAMP.value}'
                """
                execute_snowflake_statement(conn, query)

            else:
                query = f"""
                    INSERT INTO {self.registry_path}."FEAST_METADATA"
                        VALUES
                        ('{project}', '{FeastMetadataKeys.LAST_UPDATED_TIMESTAMP.value}', '{update_time}', CURRENT_TIMESTAMP())
                """
                execute_snowflake_statement(conn, query)

    def commit(self):
        pass



# Copyright 2022 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import random
import time
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, cast

import requests
from rockset.exceptions import BadRequestException, RocksetException
from rockset.models import QueryRequestSql
from rockset.query_paginator import QueryPaginator
from rockset.rockset_client import RocksetClient

from feast.entity import Entity
from feast.feature_view import FeatureView
from feast.infra.online_stores.helpers import compute_entity_id
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.usage import log_exceptions_and_usage

logger = logging.getLogger(__name__)


class RocksetOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for Rockset store"""

    type: Literal["rockset"] = "rockset"
    """Online store type selector"""

    api_key: Optional[str] = None
    """Api Key to be used for Rockset Account. If not set the env var ROCKSET_APIKEY will be used."""

    host: Optional[str] = None
    """The Host Url for Rockset requests. If not set the env var ROCKSET_APISERVER will be used."""

    read_pagination_batch_size: int = 100
    """Batch size of records that will be turned per page when paginating a batched read"""

    collection_created_timeout_secs: int = 60
    """The amount of time, in seconds, we will wait for the collection to become visible to the API"""

    collection_ready_timeout_secs: int = 30 * 60
    """The amount of time, in seconds, we will wait for the collection to enter READY state"""

    fence_all_writes: bool = True
    """Whether to wait for all writes to be flushed from log and queryable. If False, documents that are written may not be seen immediately in subsequent reads"""

    fence_timeout_secs: int = 10 * 60
    """The amount of time we will wait, in seconds, for the write fence to be passed"""

    initial_request_backoff_secs: int = 2
    """Initial backoff, in seconds, we will wait between requests when polling for a response"""

    max_request_backoff_secs: int = 30
    """Initial backoff, in seconds, we will wait between requests when polling for a response"""

    max_request_attempts: int = 10 * 1000
    """The max amount of times we will retry a failed request"""


class RocksetOnlineStore(OnlineStore):
    """
    Rockset implementation of the online store interface.

    Attributes:
        _rockset_client: Rockset openapi client.
    """

    _rockset_client = None

    @log_exceptions_and_usage(online_store="rockset")
    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Write a batch of feature rows to online Rockset store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            data: a list of quadruplets containing Feature data. Each quadruplet contains an Entity Key,
            a dict containing feature values, an event timestamp for the row, and
            the created timestamp for the row if it exists.
            progress: Optional function to be called once every mini-batch of rows is written to
            the online store. Can be used to display progress.
        """

        online_config = config.online_store
        assert isinstance(online_config, RocksetOnlineStoreConfig)

        rs = self.get_rockset_client(online_config)
        collection_name = self.get_collection_name(config, table)

        # We need to deduplicate on entity_id and we will save the latest timestamp version.
        dedup_dict = {}
        for feature_vals in data:
            entity_key, features, timestamp, created_ts = feature_vals
            serialized_key = compute_entity_id(
                entity_key=entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )

            if serialized_key not in dedup_dict:
                dedup_dict[serialized_key] = feature_vals
                continue

            # If the entity already existings in the dictionary ignore the entry if it has a lower timestamp.
            if timestamp <= dedup_dict[serialized_key][2]:
                continue

            dedup_dict[serialized_key] = feature_vals

        request_batch = []
        for serialized_key, feature_vals in dedup_dict.items():
            document = {}
            entity_key, features, timestamp, created_ts = feature_vals
            document["_id"] = serialized_key

            # Rockset python client currently does not handle datetime correctly and will convert
            # to string instead of native Rockset DATETIME. This will be fixed, but until then we
            # use isoformat.
            document["event_ts"] = timestamp.isoformat()
            document["created_ts"] = (
                "" if created_ts is None else created_ts.isoformat()
            )
            for k, v in features.items():
                # Rockset client currently does not support bytes type.
                document[k] = v.SerializeToString().hex()

            # TODO: Implement async batching with retries.
            request_batch.append(document)

            if progress:
                progress(1)

        resp = rs.Documents.add_documents(
            collection=collection_name, data=request_batch
        )
        if online_config.fence_all_writes:
            self.wait_for_fence(rs, collection_name, resp["last_offset"], online_config)

        return None

    @log_exceptions_and_usage(online_store="rockset")
    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Retrieve feature values from the online Rockset store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            entity_keys: a list of entity keys that should be read from the FeatureStore.
        """
        online_config = config.online_store
        assert isinstance(online_config, RocksetOnlineStoreConfig)

        rs = self.get_rockset_client(online_config)
        collection_name = self.get_collection_name(config, table)

        feature_list = ""
        if requested_features is not None:
            feature_list = ",".join(requested_features)

        entity_serialized_key_list = [
            compute_entity_id(
                k,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            for k in entity_keys
        ]

        entity_query_str = ",".join(
            "'{id}'".format(id=s) for s in entity_serialized_key_list
        )

        query_str = f"""
            SELECT
                "_id",
                "event_ts",
                {feature_list}
            FROM
                {collection_name}
            WHERE
                "_id" IN ({entity_query_str})
        """

        feature_set = set()
        if requested_features:
            feature_set.update(requested_features)

        result_map = {}
        for page in QueryPaginator(
            rs,
            rs.Queries.query(
                sql=QueryRequestSql(
                    query=query_str,
                    paginate=True,
                    initial_paginate_response_doc_count=online_config.read_pagination_batch_size,
                )
            ),
        ):
            for doc in page:
                result = {}
                for k, v in doc.items():
                    if k not in feature_set:
                        # We want to skip deserializing values that are not feature values like bookeeping values.
                        continue

                    val = ValueProto()

                    # TODO: Remove bytes <-> string parsing once client supports bytes.
                    val.ParseFromString(bytes.fromhex(v))
                    result[k] = val
                result_map[doc["_id"]] = (
                    datetime.fromisoformat(doc["event_ts"]),
                    result,
                )

        results_list: List[
            Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]
        ] = []
        for key in entity_serialized_key_list:
            if key not in result_map:
                # If not found, we add a gap to let the client know.
                results_list.append((None, None))
                continue

            results_list.append(result_map[key])

        return results_list

    @log_exceptions_and_usage(online_store="rockset")
    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        Update tables from the Rockset Online Store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables_to_delete: Tables to delete from the Rockset Online Store.
            tables_to_keep: Tables to keep in the Rockset Online Store.
        """
        online_config = config.online_store
        assert isinstance(online_config, RocksetOnlineStoreConfig)
        rs = self.get_rockset_client(online_config)

        created_collections = []
        for table_instance in tables_to_keep:
            try:
                collection_name = self.get_collection_name(config, table_instance)
                rs.Collections.create_file_upload_collection(name=collection_name)
                created_collections.append(collection_name)
            except BadRequestException as e:
                if self.parse_request_error_type(e) == "AlreadyExists":
                    # Table already exists nothing to do. We should still make sure it is ready though.
                    created_collections.append(collection_name)
                    continue
                raise

        for table_to_delete in tables_to_delete:
            self.delete_collection(
                rs, collection_name=self.get_collection_name(config, table_to_delete)
            )

        # Now wait for all collections to be READY.
        self.wait_for_ready_collections(
            rs, created_collections, online_config=online_config
        )

    @log_exceptions_and_usage(online_store="rockset")
    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        """
        Delete all collections from the Rockset Online Store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables: Tables to delete from the feature repo.
        """
        online_config = config.online_store
        assert isinstance(online_config, RocksetOnlineStoreConfig)
        rs = self.get_rockset_client(online_config)
        for table in tables:
            self.delete_collection(
                rs, collection_name=self.get_collection_name(config, table)
            )

    def get_rockset_client(
        self, onlineConfig: RocksetOnlineStoreConfig
    ) -> RocksetClient:
        """
        Fetches the RocksetClient to be used for all requests for this online store based on the api
        configuration in the provided config. If no configuration provided local ENV vars will be used.

        Args:
            onlineConfig: The RocksetOnlineStoreConfig associated with this online store.
        """
        if self._rockset_client is not None:
            return self._rockset_client

        _api_key = (
            os.getenv("ROCKSET_APIKEY")
            if isinstance(onlineConfig.api_key, type(None))
            else onlineConfig.api_key
        )
        _host = (
            os.getenv("ROCKSET_APISERVER")
            if isinstance(onlineConfig.host, type(None))
            else onlineConfig.host
        )
        self._rockset_client = RocksetClient(host=_host, api_key=_api_key)
        return self._rockset_client

    @staticmethod
    def delete_collection(rs: RocksetClient, collection_name: str):
        """
        Deletes the collection  whose name was provided

        Args:
            rs: The RocksetClient to be used for the deletion.
            collection_name: The name of the collection to be deleted.
        """

        try:
            rs.Collections.delete(collection=collection_name)
        except RocksetException as e:
            if RocksetOnlineStore.parse_request_error_type(e) == "NotFound":
                logger.warning(
                    f"Trying to delete collection that does not exist {collection_name}"
                )
                return
            raise

    @staticmethod
    def get_collection_name(config: RepoConfig, feature_view: FeatureView) -> str:
        """
        Returns the collection name based on the provided config and FeatureView.

        Args:
            config: RepoConfig for the online store.
            feature_view: FeatureView that is backed by the returned collection name.

        Returns:
            The collection name as a string.
        """
        project_val = config.project if config.project else "feast"
        table_name = feature_view.name if feature_view.name else "feature_store"
        return f"{project_val}_{table_name}"

    @staticmethod
    def parse_request_error_type(e: RocksetException) -> str:
        """
        Parse a throw RocksetException. Will return a string representing the type of error that was thrown.

        Args:
            e: The RockException that is being parsed.

        Returns:
            Error type parsed as a string.
        """

        body_dict = json.loads(e.body)
        return body_dict["type"]

    @staticmethod
    def wait_for_fence(
        rs: RocksetClient,
        collection_name: str,
        last_offset: str,
        online_config: RocksetOnlineStoreConfig,
    ):
        """
        Waits until 'last_offset' is flushed and values are ready to be read. If wait lasts longer than the timeout specified in config
        a timeout exception will be throw.

        Args:
            rs: Rockset client that will be used to make all requests.
            collection_name: Collection associated with the offsets we are waiting for.
            last_offset: The actual offsets we are waiting to be flushed.
            online_config: The config that will be used to determine timeouts and backout configurations.
        """

        resource_path = (
            f"/v1/orgs/self/ws/commons/collections/{collection_name}/offsets/commit"
        )
        request = {"name": [last_offset]}

        headers = {}
        headers["Content-Type"] = "application/json"
        headers["Authorization"] = f"ApiKey {rs.api_client.configuration.api_key}"

        t_start = time.time()
        for num_attempts in range(online_config.max_request_attempts):
            delay = time.time() - t_start
            resp = requests.post(
                url=f"{rs.api_client.configuration.host}{resource_path}",
                json=request,
                headers=headers,
            )

            if resp.status_code == 200 and resp.json()["data"]["passed"] is True:
                break

            if delay > online_config.fence_timeout_secs:
                raise TimeoutError(
                    f"Write to collection {collection_name} at offset {last_offset} was not available for read after {delay} secs"
                )

            if resp.status_code == 429:
                RocksetOnlineStore.backoff_sleep(num_attempts, online_config)
                continue
            elif resp.status_code != 200:
                raise Exception(f"[{resp.status_code}]: {resp.reason}")

            RocksetOnlineStore.backoff_sleep(num_attempts, online_config)

    @staticmethod
    def wait_for_ready_collections(
        rs: RocksetClient,
        collection_names: List[str],
        online_config: RocksetOnlineStoreConfig,
    ):
        """
        Waits until all collections provided have entered READY state and can accept new documents. If wait
        lasts longer than timeout a TimeoutError exception will be thrown.

        Args:
            rs: Rockset client that will be used to make all requests.
            collection_names: All collections that we will wait for.
            timeout: The max amount of time we will wait for the collections to become READY.
        """

        t_start = time.time()
        for cname in collection_names:
            # We will wait until the provided timeout for all collections to become READY.
            for num_attempts in range(online_config.max_request_attempts):
                resp = None
                delay = time.time() - t_start
                try:
                    resp = rs.Collections.get(collection=cname)
                except RocksetException as e:
                    error_type = RocksetOnlineStore.parse_request_error_type(e)
                    if error_type == "NotFound":
                        if delay > online_config.collection_created_timeout_secs:
                            raise TimeoutError(
                                f"Collection {cname} failed to become visible after {delay} seconds"
                            )
                    elif error_type == "RateLimitExceeded":
                        RocksetOnlineStore.backoff_sleep(num_attempts, online_config)
                        continue
                    else:
                        raise

                if (
                    resp is not None
                    and cast(Dict[str, dict], resp)["data"]["status"] == "READY"
                ):
                    break

                if delay > online_config.collection_ready_timeout_secs:
                    raise TimeoutError(
                        f"Collection {cname} failed to become ready after {delay} seconds"
                    )

                RocksetOnlineStore.backoff_sleep(num_attempts, online_config)

    @staticmethod
    def backoff_sleep(attempts: int, online_config: RocksetOnlineStoreConfig):
        """
        Sleep for the needed amount of time based on the number of request attempts.

        Args:
            backoff: The amount of time we will sleep for
            max_backoff: The max amount of time we should ever backoff for.
            rate_limited: Whether this method is being called as part of a rate limited request.
        """

        default_backoff = online_config.initial_request_backoff_secs

        # Full jitter, exponential backoff.
        backoff = random.uniform(
            default_backoff,
            min(default_backoff << attempts, online_config.max_request_backoff_secs),
        )
        time.sleep(backoff)



import contextlib
import uuid
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import pyarrow
import pyarrow as pa
from dateutil import parser
from pydantic import StrictStr, model_validator
from pytz import utc

from feast import OnDemandFeatureView, RedshiftSource
from feast.data_source import DataSource
from feast.errors import InvalidEntityType
from feast.feature_logging import LoggingConfig, LoggingSource
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.redshift_source import (
    RedshiftLoggingDestination,
    SavedDatasetRedshiftStorage,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.utils import aws_utils
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.usage import log_exceptions_and_usage


class RedshiftOfflineStoreConfig(FeastConfigBaseModel):
    """Offline store config for AWS Redshift"""

    type: Literal["redshift"] = "redshift"
    """ Offline store type selector"""

    cluster_id: Optional[StrictStr] = None
    """ Redshift cluster identifier, for provisioned clusters """

    user: Optional[StrictStr] = None
    """ Redshift user name, only required for provisioned clusters """

    workgroup: Optional[StrictStr] = None
    """ Redshift workgroup identifier, for serverless """

    region: StrictStr
    """ Redshift cluster's AWS region """

    database: StrictStr
    """ Redshift database name """

    s3_staging_location: StrictStr
    """ S3 path for importing & exporting data to Redshift """

    iam_role: StrictStr
    """ IAM Role for Redshift, granting it access to S3 """

    @model_validator(mode="after")
    def require_cluster_and_user_or_workgroup(self):
        """
        Provisioned Redshift clusters:  Require cluster_id and user, ignore workgroup
        Serverless Redshift:  Require workgroup, ignore cluster_id and user
        """
        cluster_id, user, workgroup = (
            self.cluster_id,
            self.user,
            self.workgroup,
        )
        if not (cluster_id and user) and not workgroup:
            raise ValueError(
                "please specify either cluster_id & user if using provisioned clusters, or workgroup if using serverless"
            )
        elif cluster_id and workgroup:
            raise ValueError("cannot specify both cluster_id and workgroup")

        return self


class RedshiftOfflineStore(OfflineStore):
    @staticmethod
    @log_exceptions_and_usage(offline_store="redshift")
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(data_source, RedshiftSource)
        assert isinstance(config.offline_store, RedshiftOfflineStoreConfig)

        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(join_key_columns)
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamp_columns = [timestamp_field]
        if created_timestamp_column:
            timestamp_columns.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(timestamp_columns) + " DESC"
        field_string = ", ".join(
            join_key_columns + feature_name_columns + timestamp_columns
        )

        redshift_client = aws_utils.get_redshift_data_client(
            config.offline_store.region
        )
        s3_resource = aws_utils.get_s3_resource(config.offline_store.region)

        start_date = start_date.astimezone(tz=utc)
        end_date = end_date.astimezone(tz=utc)

        query = f"""
            SELECT
                {field_string}
                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
            FROM (
                SELECT {field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression}
                WHERE {timestamp_field} BETWEEN TIMESTAMP '{start_date}' AND TIMESTAMP '{end_date}'
            )
            WHERE _feast_row = 1
            """
        # When materializing a single feature view, we don't need full feature names. On demand transforms aren't materialized
        return RedshiftRetrievalJob(
            query=query,
            redshift_client=redshift_client,
            s3_resource=s3_resource,
            config=config,
            full_feature_names=False,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="redshift")
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, RedshiftOfflineStoreConfig)
        assert isinstance(data_source, RedshiftSource)
        from_expression = data_source.get_table_query_string()

        field_string = ", ".join(
            join_key_columns + feature_name_columns + [timestamp_field]
        )

        redshift_client = aws_utils.get_redshift_data_client(
            config.offline_store.region
        )
        s3_resource = aws_utils.get_s3_resource(config.offline_store.region)

        start_date = start_date.astimezone(tz=utc)
        end_date = end_date.astimezone(tz=utc)

        query = f"""
            SELECT {field_string}
            FROM {from_expression}
            WHERE {timestamp_field} BETWEEN TIMESTAMP '{start_date}' AND TIMESTAMP '{end_date}'
        """

        return RedshiftRetrievalJob(
            query=query,
            redshift_client=redshift_client,
            s3_resource=s3_resource,
            config=config,
            full_feature_names=False,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="redshift")
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, RedshiftOfflineStoreConfig)
        for fv in feature_views:
            assert isinstance(fv.batch_source, RedshiftSource)

        redshift_client = aws_utils.get_redshift_data_client(
            config.offline_store.region
        )
        s3_resource = aws_utils.get_s3_resource(config.offline_store.region)

        entity_schema = _get_entity_schema(
            entity_df, redshift_client, config, s3_resource
        )

        entity_df_event_timestamp_col = (
            offline_utils.infer_event_timestamp_from_entity_df(entity_schema)
        )

        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df,
            entity_df_event_timestamp_col,
            redshift_client,
            config,
        )

        @contextlib.contextmanager
        def query_generator() -> Iterator[str]:
            table_name = offline_utils.get_temp_entity_table_name()

            _upload_entity_df(
                entity_df, redshift_client, config, s3_resource, table_name
            )

            expected_join_keys = offline_utils.get_expected_join_keys(
                project, feature_views, registry
            )

            offline_utils.assert_expected_columns_in_entity_df(
                entity_schema, expected_join_keys, entity_df_event_timestamp_col
            )

            # Build a query context containing all information required to template the Redshift SQL query
            query_context = offline_utils.get_feature_view_query_context(
                feature_refs,
                feature_views,
                registry,
                project,
                entity_df_event_timestamp_range,
            )

            # Generate the Redshift SQL query from the query context
            query = offline_utils.build_point_in_time_query(
                query_context,
                left_table_query_string=table_name,
                entity_df_event_timestamp_col=entity_df_event_timestamp_col,
                entity_df_columns=entity_schema.keys(),
                query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                full_feature_names=full_feature_names,
            )

            try:
                yield query
            finally:
                # Always clean up the uploaded Redshift table
                aws_utils.execute_redshift_statement(
                    redshift_client,
                    config.offline_store.cluster_id,
                    config.offline_store.workgroup,
                    config.offline_store.database,
                    config.offline_store.user,
                    f"DROP TABLE IF EXISTS {table_name}",
                )

        return RedshiftRetrievalJob(
            query=query_generator,
            redshift_client=redshift_client,
            s3_resource=s3_resource,
            config=config,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(entity_schema.keys() - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
        )

    @staticmethod
    def write_logged_features(
        config: RepoConfig,
        data: Union[pyarrow.Table, Path],
        source: LoggingSource,
        logging_config: LoggingConfig,
        registry: BaseRegistry,
    ):
        destination = logging_config.destination
        assert isinstance(destination, RedshiftLoggingDestination)

        redshift_client = aws_utils.get_redshift_data_client(
            config.offline_store.region
        )
        s3_resource = aws_utils.get_s3_resource(config.offline_store.region)
        if isinstance(data, Path):
            s3_path = f"{config.offline_store.s3_staging_location}/logged_features/{uuid.uuid4()}"
        else:
            s3_path = f"{config.offline_store.s3_staging_location}/logged_features/{uuid.uuid4()}.parquet"

        aws_utils.upload_arrow_table_to_redshift(
            table=data,
            redshift_data_client=redshift_client,
            cluster_id=config.offline_store.cluster_id,
            workgroup=config.offline_store.workgroup,
            database=config.offline_store.database,
            user=config.offline_store.user,
            s3_resource=s3_resource,
            s3_path=s3_path,
            iam_role=config.offline_store.iam_role,
            table_name=destination.table_name,
            schema=source.get_schema(registry),
            fail_if_exists=False,
        )

    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pyarrow.Table,
        progress: Optional[Callable[[int], Any]],
    ):
        assert isinstance(config.offline_store, RedshiftOfflineStoreConfig)
        assert isinstance(feature_view.batch_source, RedshiftSource)

        pa_schema, column_names = offline_utils.get_pyarrow_schema_from_batch_source(
            config, feature_view.batch_source
        )
        if column_names != table.column_names:
            raise ValueError(
                f"The input pyarrow table has schema {table.schema} with the incorrect columns {table.column_names}. "
                f"The schema is expected to be {pa_schema} with the columns (in this exact order) to be {column_names}."
            )

        if table.schema != pa_schema:
            table = table.cast(pa_schema)

        redshift_options = feature_view.batch_source.redshift_options
        redshift_client = aws_utils.get_redshift_data_client(
            config.offline_store.region
        )
        s3_resource = aws_utils.get_s3_resource(config.offline_store.region)

        aws_utils.upload_arrow_table_to_redshift(
            table=table,
            redshift_data_client=redshift_client,
            cluster_id=config.offline_store.cluster_id,
            workgroup=config.offline_store.workgroup,
            database=redshift_options.database
            # Users can define database in the source if needed but it's not required.
            or config.offline_store.database,
            user=config.offline_store.user,
            s3_resource=s3_resource,
            s3_path=f"{config.offline_store.s3_staging_location}/push/{uuid.uuid4()}.parquet",
            iam_role=config.offline_store.iam_role,
            table_name=redshift_options.fully_qualified_table_name,
            schema=pa_schema,
            fail_if_exists=False,
        )


class RedshiftRetrievalJob(RetrievalJob):
    def __init__(
        self,
        query: Union[str, Callable[[], ContextManager[str]]],
        redshift_client,
        s3_resource,
        config: RepoConfig,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        """Initialize RedshiftRetrievalJob object.

        Args:
            query: Redshift SQL query to execute. Either a string, or a generator function that handles the artifact cleanup.
            redshift_client: boto3 redshift-data client
            s3_resource: boto3 s3 resource object
            config: Feast repo config
            full_feature_names: Whether to add the feature view prefixes to the feature names
            on_demand_feature_views (optional): A list of on demand transforms to apply at retrieval time
        """
        if not isinstance(query, str):
            self._query_generator = query
        else:

            @contextlib.contextmanager
            def query_generator() -> Iterator[str]:
                assert isinstance(query, str)
                yield query

            self._query_generator = query_generator
        self._redshift_client = redshift_client
        self._s3_resource = s3_resource
        self._config = config
        self._s3_path = (
            self._config.offline_store.s3_staging_location
            + "/unload/"
            + str(uuid.uuid4())
        )
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    @log_exceptions_and_usage
    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        with self._query_generator() as query:
            return aws_utils.unload_redshift_query_to_df(
                self._redshift_client,
                self._config.offline_store.cluster_id,
                self._config.offline_store.workgroup,
                self._config.offline_store.database,
                self._config.offline_store.user,
                self._s3_resource,
                self._s3_path,
                self._config.offline_store.iam_role,
                query,
            )

    @log_exceptions_and_usage
    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        with self._query_generator() as query:
            return aws_utils.unload_redshift_query_to_pa(
                self._redshift_client,
                self._config.offline_store.cluster_id,
                self._config.offline_store.workgroup,
                self._config.offline_store.database,
                self._config.offline_store.user,
                self._s3_resource,
                self._s3_path,
                self._config.offline_store.iam_role,
                query,
            )

    @log_exceptions_and_usage
    def to_s3(self) -> str:
        """Export dataset to S3 in Parquet format and return path"""
        if self.on_demand_feature_views:
            transformed_df = self.to_df()
            aws_utils.upload_df_to_s3(self._s3_resource, self._s3_path, transformed_df)
            return self._s3_path

        with self._query_generator() as query:
            aws_utils.execute_redshift_query_and_unload_to_s3(
                self._redshift_client,
                self._config.offline_store.cluster_id,
                self._config.offline_store.workgroup,
                self._config.offline_store.database,
                self._config.offline_store.user,
                self._s3_path,
                self._config.offline_store.iam_role,
                query,
            )
            return self._s3_path

    @log_exceptions_and_usage
    def to_redshift(self, table_name: str) -> None:
        """Save dataset as a new Redshift table"""
        if self.on_demand_feature_views:
            transformed_df = self.to_df()
            aws_utils.upload_df_to_redshift(
                self._redshift_client,
                self._config.offline_store.cluster_id,
                self._config.offline_store.workgroup,
                self._config.offline_store.database,
                self._config.offline_store.user,
                self._s3_resource,
                f"{self._config.offline_store.s3_staging_location}/features_df/{table_name}.parquet",
                self._config.offline_store.iam_role,
                table_name,
                transformed_df,
            )
            return

        with self._query_generator() as query:
            query = f'CREATE TABLE "{table_name}" AS ({query});\n'

            aws_utils.execute_redshift_statement(
                self._redshift_client,
                self._config.offline_store.cluster_id,
                self._config.offline_store.workgroup,
                self._config.offline_store.database,
                self._config.offline_store.user,
                query,
            )

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: Optional[bool] = False,
        timeout: Optional[int] = None,
    ):
        assert isinstance(storage, SavedDatasetRedshiftStorage)
        self.to_redshift(table_name=storage.redshift_options.table)

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def supports_remote_storage_export(self) -> bool:
        return True

    def to_remote_storage(self) -> List[str]:
        path = self.to_s3()
        return aws_utils.list_s3_files(self._config.offline_store.region, path)


def _upload_entity_df(
    entity_df: Union[pd.DataFrame, str],
    redshift_client,
    config: RepoConfig,
    s3_resource,
    table_name: str,
):
    if isinstance(entity_df, pd.DataFrame):
        # If the entity_df is a pandas dataframe, upload it to Redshift
        aws_utils.upload_df_to_redshift(
            redshift_client,
            config.offline_store.cluster_id,
            config.offline_store.workgroup,
            config.offline_store.database,
            config.offline_store.user,
            s3_resource,
            f"{config.offline_store.s3_staging_location}/entity_df/{table_name}.parquet",
            config.offline_store.iam_role,
            table_name,
            entity_df,
        )
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), create a Redshift table out of it
        aws_utils.execute_redshift_statement(
            redshift_client,
            config.offline_store.cluster_id,
            config.offline_store.workgroup,
            config.offline_store.database,
            config.offline_store.user,
            f"CREATE TABLE {table_name} AS ({entity_df})",
        )
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_schema(
    entity_df: Union[pd.DataFrame, str],
    redshift_client,
    config: RepoConfig,
    s3_resource,
) -> Dict[str, np.dtype]:
    if isinstance(entity_df, pd.DataFrame):
        return dict(zip(entity_df.columns, entity_df.dtypes))

    elif isinstance(entity_df, str):
        # get pandas dataframe consisting of 1 row (LIMIT 1) and generate the schema out of it
        entity_df_sample = RedshiftRetrievalJob(
            f"SELECT * FROM ({entity_df}) LIMIT 1",
            redshift_client,
            s3_resource,
            config,
            full_feature_names=False,
        ).to_df()
        return dict(zip(entity_df_sample.columns, entity_df_sample.dtypes))
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pd.DataFrame, str],
    entity_df_event_timestamp_col: str,
    redshift_client,
    config: RepoConfig,
) -> Tuple[datetime, datetime]:
    if isinstance(entity_df, pd.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pd.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pd.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), determine range
        # from table
        statement_id = aws_utils.execute_redshift_statement(
            redshift_client,
            config.offline_store.cluster_id,
            config.offline_store.workgroup,
            config.offline_store.database,
            config.offline_store.user,
            f"SELECT MIN({entity_df_event_timestamp_col}) AS min, MAX({entity_df_event_timestamp_col}) AS max "
            f"FROM ({entity_df})",
        )
        res = aws_utils.get_redshift_statement_result(redshift_client, statement_id)[
            "Records"
        ][0]
        entity_df_event_timestamp_range = (
            parser.parse(res[0]["stringValue"]),
            parser.parse(res[1]["stringValue"]),
        )
    else:
        raise InvalidEntityType(type(entity_df))

    return entity_df_event_timestamp_range


# This query is based on sdk/python/feast/infra/offline_stores/bigquery.py:MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN
# There are couple of changes from BigQuery:
# 1. Use VARCHAR instead of STRING type
# 2. Use "t - x * interval '1' second" instead of "Timestamp_sub(...)"
# 3. Replace `SELECT * EXCEPT (...)` with `SELECT *`, because `EXCEPT` is not supported by Redshift.
#    Instead, we drop the column later after creating the table out of the query.
# We need to keep this query in sync with BigQuery.

MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
/*
 Compute a deterministic hash for the `left_table_query_string` that will be used throughout
 all the logic as the field to GROUP BY the data
*/
WITH entity_dataframe AS (
    SELECT *,
        {{entity_df_event_timestamp_col}} AS entity_timestamp
        {% for featureview in featureviews %}
            {% if featureview.entities %}
            ,(
                {% for entity in featureview.entities %}
                    CAST({{entity}} as VARCHAR) ||
                {% endfor %}
                CAST({{entity_df_event_timestamp_col}} AS VARCHAR)
            ) AS {{featureview.name}}__entity_row_unique_id
            {% else %}
            ,CAST({{entity_df_event_timestamp_col}} AS VARCHAR) AS {{featureview.name}}__entity_row_unique_id
            {% endif %}
        {% endfor %}
    FROM {{ left_table_query_string }}
),

{% for featureview in featureviews %}

{{ featureview.name }}__entity_dataframe AS (
    SELECT
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
    FROM entity_dataframe
    GROUP BY
        {{ featureview.entities | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        entity_timestamp,
        {{featureview.name}}__entity_row_unique_id
),

/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.

 1. We first join the current feature_view to the entity dataframe that has been passed.
 This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the `timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously

 The output of this CTE will contain all the necessary information and already filtered out most
 of the data that is not relevant.
*/

{{ featureview.name }}__subquery AS (
    SELECT
        {{ featureview.timestamp_field }} as event_timestamp,
        {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
        {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
        {% for feature in featureview.features %}
            {{ feature }} as {% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}{% if loop.last %}{% else %}, {% endif %}
        {% endfor %}
    FROM {{ featureview.table_subquery }}
    WHERE {{ featureview.timestamp_field }} <= '{{ featureview.max_event_timestamp }}'
    {% if featureview.ttl == 0 %}{% else %}
    AND {{ featureview.timestamp_field }} >= '{{ featureview.min_event_timestamp }}'
    {% endif %}
),

{{ featureview.name }}__base AS (
    SELECT
        subquery.*,
        entity_dataframe.entity_timestamp,
        entity_dataframe.{{featureview.name}}__entity_row_unique_id
    FROM {{ featureview.name }}__subquery AS subquery
    INNER JOIN {{ featureview.name }}__entity_dataframe AS entity_dataframe
    ON TRUE
        AND subquery.event_timestamp <= entity_dataframe.entity_timestamp

        {% if featureview.ttl == 0 %}{% else %}
        AND subquery.event_timestamp >= entity_dataframe.entity_timestamp - {{ featureview.ttl }} * interval '1' second
        {% endif %}

        {% for entity in featureview.entities %}
        AND subquery.{{ entity }} = entity_dataframe.{{ entity }}
        {% endfor %}
),

/*
 2. If the `created_timestamp_column` has been set, we need to
 deduplicate the data first. This is done by calculating the
 `MAX(created_at_timestamp)` for each event_timestamp.
 We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
{{ featureview.name }}__dedup AS (
    SELECT
        {{featureview.name}}__entity_row_unique_id,
        event_timestamp,
        MAX(created_timestamp) as created_timestamp
    FROM {{ featureview.name }}__base
    GROUP BY {{featureview.name}}__entity_row_unique_id, event_timestamp
),
{% endif %}

/*
 3. The data has been filtered during the first CTE "*__base"
 Thus we only need to compute the latest timestamp of each feature.
*/
{{ featureview.name }}__latest AS (
    SELECT
        event_timestamp,
        {% if featureview.created_timestamp_column %}created_timestamp,{% endif %}
        {{featureview.name}}__entity_row_unique_id
    FROM
    (
        SELECT *,
            ROW_NUMBER() OVER(
                PARTITION BY {{featureview.name}}__entity_row_unique_id
                ORDER BY event_timestamp DESC{% if featureview.created_timestamp_column %},created_timestamp DESC{% endif %}
            ) AS row_number
        FROM {{ featureview.name }}__base
        {% if featureview.created_timestamp_column %}
            INNER JOIN {{ featureview.name }}__dedup
            USING ({{featureview.name}}__entity_row_unique_id, event_timestamp, created_timestamp)
        {% endif %}
    )
    WHERE row_number = 1
),

/*
 4. Once we know the latest value of each feature for a given timestamp,
 we can join again the data back to the original "base" dataset
*/
{{ featureview.name }}__cleaned AS (
    SELECT base.*
    FROM {{ featureview.name }}__base as base
    INNER JOIN {{ featureview.name }}__latest
    USING(
        {{featureview.name}}__entity_row_unique_id,
        event_timestamp
        {% if featureview.created_timestamp_column %}
            ,created_timestamp
        {% endif %}
    )
){% if loop.last %}{% else %}, {% endif %}


{% endfor %}
/*
 Joins the outputs of multiple time travel joins to a single table.
 The entity_dataframe dataset being our source of truth here.
 */

SELECT {{ final_output_feature_names | join(', ')}}
FROM entity_dataframe
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
        {{featureview.name}}__entity_row_unique_id
        {% for feature in featureview.features %}
            ,{% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}
        {% endfor %}
    FROM {{ featureview.name }}__cleaned
) USING ({{featureview.name}}__entity_row_unique_id)
{% endfor %}
"""



from typing import Dict

from feast.repo_config import FeastConfigBaseModel
from feast.type_map import feast_value_type_to_pa
from feast.value_type import ValueType
from pydantic import Field, StrictStr


class DorisSQLConfig(FeastConfigBaseModel):
    host: StrictStr = Field(default="127.0.0.1")
    port: int = 9030
    database: StrictStr = Field(default="demo")
    user: StrictStr = Field(default="root")
    password: StrictStr = Field(default="root")


def doris_type_code_to_arrow(code: int) -> str:
    return feast_value_type_to_pa(
        doris_type_to_feast_value_type(doris_type_code_to_doris_type(code))
    )


def arrow_to_doris_type(t_str: str) -> str:
    try:
        if t_str.startswith("timestamp") or t_str.startswith("datetime"):
            return "DATETIME" if "tz=" in t_str else "DATETIME"
        return {
            "null": "NULL",
            "bool": "BOOLEAN",
            "int8": "TINYINT",
            "int16": "SMALLINT",
            "int32": "INT",
            "int64": "BIGINT",
            "list<item: int32>": "ARRAY<INT>",
            "list<item: int64>": "ARRAY<BIGINT>",
            "list<item: bool>": "ARRAY<BOOLEAN>",
            "list<item: double>": "ARRAY<DOUBLE>",
            "list<item: timestamp[us]>": "ARRAY<DATETIME>",
            "uint8": "SMALLINT",
            "uint16": "INT",
            "uint32": "BIGINT",
            "uint64": "BIGINT",
            "float": "FLOAT",
            "double": "DOUBLE",
            "binary": "BINARY",
            "string": "STRING",
        }[t_str]
    except KeyError:
        raise ValueError(f"Unsupported type: {t_str}")


def doris_type_code_to_doris_type(code: int) -> str:
    DORIS_TYPE_MAP = {
        1: "boolean",
        2: "smallint",
        3: "int",
        8: "bigint",
        254: "struct",
        4: "float",
        5: "double",
        246: "decimal",
        10: "date",
        12: "datetime",
        144: "map",
        252: "string",
    }

    return DORIS_TYPE_MAP.get(code, "string")


def doris_type_to_feast_value_type(type_str: str) -> ValueType:
    type_map: Dict[str, ValueType] = {
        "boolean": ValueType.BOOL,
        "smallint": ValueType.INT32,
        "int": ValueType.INT32,
        "bigint": ValueType.INT64,
        "float": ValueType.FLOAT,
        "double": ValueType.DOUBLE,
        "decimal": ValueType.DOUBLE,
        "date": ValueType.UNIX_TIMESTAMP,
        "datetime": ValueType.UNIX_TIMESTAMP,
        "string": ValueType.STRING,
        "struct": ValueType.STRING,
        "map": ValueType.STRING,
    }
    value = (
        type_map[type_str.lower()]
        if type_str.lower() in type_map
        else ValueType.UNKNOWN
    )
    if value == ValueType.UNKNOWN:
        print("unknown type:", type_str)
    return value



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import os
import pathlib
import re
import shutil
import subprocess
import sys
from distutils.cmd import Command
from pathlib import Path

from setuptools import find_packages

try:
    from setuptools import setup
    from setuptools.command.build_ext import build_ext as _build_ext
    from setuptools.command.build_py import build_py
    from setuptools.command.develop import develop
    from setuptools.command.install import install

except ImportError:
    from distutils.command.build_py import build_py
    from distutils.core import setup

NAME = "feast-doris"
DESCRIPTION = "Python SDK for Feast"
URL = "https://github.com/feast-dev/feast"
AUTHOR = "DQLabs"
REQUIRES_PYTHON = ">=3.9.0"
VERSION = '0.1.2'
REQUIRED = [
    "click>=7.0.0,<9.0.0",
    "colorama>=0.3.9,<1",
    "dill~=0.3.0",
    "mypy-protobuf>=3.1",
    "Jinja2>=2,<4",
    "jsonschema",
    "mmh3",
    "numpy>=1.22,<1.25",
    "pandas>=1.4.3,<3",
    # Higher than 4.23.4 seems to cause a seg fault
    "protobuf>=4.24.0,<5.0.0",
    "pyarrow>=4",
    "pydantic>=2.0.0",
    "pygments>=2.12.0,<3",
    "PyYAML>=5.4.0,<7",
    "requests",
    "SQLAlchemy[mypy]>1",
    "tabulate>=0.8.0,<1",
    "tenacity>=7,<9",
    "toml>=0.10.0,<1",
    "tqdm>=4,<5",
    "typeguard>=4.0.0",
    "fastapi>=0.68.0",
    "uvicorn[standard]>=0.14.0,<1",
    "gunicorn; platform_system != 'Windows'",
    "dask==2024.6.2",
    "bowler",  # Needed for automatic repo upgrades
    # FastAPI does not correctly pull starlette dependency on httpx see thread(https://github.com/tiangolo/fastapi/issues/5656).
    "httpx>=0.23.3",
    "importlib-resources>=6.0.0,<7",
    "importlib_metadata>=6.8.0,<7",
]

GCP_REQUIRED = [
    "google-api-core>=1.23.0,<3",
    "googleapis-common-protos>=1.52.0,<2",
    "google-cloud-bigquery[pandas]>=2,<3.13.0",
    "google-cloud-bigquery-storage >= 2.0.0,<3",
    "google-cloud-datastore>=2.1.0,<3",
    "google-cloud-storage>=1.34.0,<3",
    "google-cloud-bigtable>=2.11.0,<3",
    "fsspec<=2024.1.0",
]

REDIS_REQUIRED = [
    "redis>=4.2.2,<5",
    "hiredis>=2.0.0,<3",
]

AWS_REQUIRED = ["boto3>=1.17.0,<2", "docker>=5.0.2", "fsspec<=2024.1.0"]

BYTEWAX_REQUIRED = ["bytewax==0.15.1", "docker>=5.0.2", "kubernetes<=20.13.0"]

KUBERNETES_REQUIRED = ["kubernetes<=20.13.0"]

SNOWFLAKE_REQUIRED = [
    "snowflake-connector-python[pandas]>=3.7,<4",
]

SPARK_REQUIRED = [
    "pyspark>=3.0.0,<4",
]

TRINO_REQUIRED = ["trino>=0.305.0,<0.400.0", "regex"]

POSTGRES_REQUIRED = [
    "psycopg2-binary>=2.8.3,<3",
]

MYSQL_REQUIRED = ["pymysql", "types-PyMySQL", "mysql-connector-python"]

HBASE_REQUIRED = [
    "happybase>=1.2.0,<3",
]

CASSANDRA_REQUIRED = [
    "cassandra-driver>=3.24.0,<4",
]

GE_REQUIRED = ["great_expectations>=0.15.41"]

AZURE_REQUIRED = [
    "azure-storage-blob>=0.37.0",
    "azure-identity>=1.6.1",
    "SQLAlchemy>=1.4.19",
    "pyodbc>=4.0.30",
    "pymssql",
]

ROCKSET_REQUIRED = [
    "rockset>=1.0.3",
]

HAZELCAST_REQUIRED = [
    "hazelcast-python-client>=5.1",
]

IBIS_REQUIRED = [
    "ibis-framework",
    "ibis-substrait",
]

GRPCIO_REQUIRED = [
    "grpcio>=1.56.2,<2",
    "grpcio-tools>=1.56.2,<2",
    "grpcio-reflection>=1.56.2,<2",
    "grpcio-health-checking>=1.56.2,<2",
]

DUCKDB_REQUIRED = ["ibis-framework[duckdb]"]

CI_REQUIRED = (
    [
        "build",
        "virtualenv==20.23.0",
        "cryptography>=35.0,<43",
        "ruff>=0.3.3",
        "grpcio-testing>=1.56.2,<2",
        # FastAPI does not correctly pull starlette dependency on httpx see thread(https://github.com/tiangolo/fastapi/issues/5656).
        "httpx>=0.23.3",
        "minio==7.1.0",
        "mock==2.0.0",
        "moto<5",
        "mypy>=1.4.1",
        "urllib3>=1.25.4,<3",
        "psutil==5.9.0",
        "py>=1.11.0",  # https://github.com/pytest-dev/pytest/issues/10420
        "pytest>=6.0.0,<8",
        "pytest-cov",
        "pytest-xdist",
        "pytest-benchmark>=3.4.1,<4",
        "pytest-lazy-fixture==0.6.3",
        "pytest-timeout==1.4.2",
        "pytest-ordering~=0.6.0",
        "pytest-mock==1.10.4",
        "pytest-env",
        "Sphinx>4.0.0,<7",
        "testcontainers==4.3.3",
        "firebase-admin>=5.2.0,<6",
        "pre-commit<3.3.2",
        "assertpy==1.1",
        "pip-tools",
        "pybindgen",
        "types-protobuf~=3.19.22",
        "types-python-dateutil",
        "types-pytz",
        "types-PyYAML",
        "types-redis",
        "types-requests<2.31.0",
        "types-setuptools",
        "types-tabulate",
        "virtualenv<20.24.2",
    ]
    + GCP_REQUIRED
    + REDIS_REQUIRED
    + AWS_REQUIRED
    + BYTEWAX_REQUIRED
    + KUBERNETES_REQUIRED
    + SNOWFLAKE_REQUIRED
    + SPARK_REQUIRED
    + POSTGRES_REQUIRED
    + MYSQL_REQUIRED
    + TRINO_REQUIRED
    + GE_REQUIRED
    + HBASE_REQUIRED
    + CASSANDRA_REQUIRED
    + AZURE_REQUIRED
    + ROCKSET_REQUIRED
    + HAZELCAST_REQUIRED
    + IBIS_REQUIRED
    + GRPCIO_REQUIRED
    + DUCKDB_REQUIRED
)

DOCS_REQUIRED = CI_REQUIRED
DEV_REQUIRED = CI_REQUIRED

# Get git repo root directory
repo_root = str(pathlib.Path(__file__).resolve().parent)

# README file from Feast repo root directory
README_FILE = os.path.join(repo_root, "README.md")
with open(README_FILE, "r", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

# Add Support for parsing tags that have a prefix containing '/' (ie 'sdk/go') to setuptools_scm.
# Regex modified from default tag regex in:
# https://github.com/pypa/setuptools_scm/blob/2a1b46d38fb2b8aeac09853e660bcd0d7c1bc7be/src/setuptools_scm/config.py#L9
TAG_REGEX = re.compile(
    r"^(?:[\/\w-]+)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]*)(?:\+.*)?$"
)

# Only set use_scm_version if git executable exists (setting this variable causes pip to use git under the hood)
if shutil.which("git"):
    use_scm_version = {"root": ".", "relative_to": __file__, "tag_regex": TAG_REGEX}
else:
    use_scm_version = None

PROTO_SUBDIRS = ["core", "registry", "serving", "types", "storage"]
PYTHON_CODE_PREFIX = "sdk/python"


class BuildPythonProtosCommand(Command):
    description = "Builds the proto files into Python files."
    user_options = [
        ("inplace", "i", "Write generated proto files to source directory."),
    ]

    def initialize_options(self):
        self.python_protoc = [
            sys.executable,
            "-m",
            "grpc_tools.protoc",
        ]  # find_executable("protoc")
        self.proto_folder = os.path.join(repo_root, "protos")
        self.sub_folders = PROTO_SUBDIRS
        self.build_lib = None
        self.inplace = 0

    def finalize_options(self):
        self.set_undefined_options("build", ("build_lib", "build_lib"))

    @property
    def python_folder(self):
        if self.inplace:
            return os.path.join(
                os.path.dirname(__file__) or os.getcwd(), "sdk/python/feast/protos"
            )

        return os.path.join(self.build_lib, "feast/protos")

    def _generate_python_protos(self, path: str):
        proto_files = glob.glob(os.path.join(self.proto_folder, path))
        Path(self.python_folder).mkdir(parents=True, exist_ok=True)
        subprocess.check_call(
            self.python_protoc
            + [
                "-I",
                self.proto_folder,
                "--python_out",
                self.python_folder,
                "--grpc_python_out",
                self.python_folder,
                "--mypy_out",
                self.python_folder,
            ]
            + proto_files
        )

    def run(self):
        for sub_folder in self.sub_folders:
            self._generate_python_protos(f"feast/{sub_folder}/*.proto")
            # We need the __init__ files for each of the generated subdirs
            # so that they are regular packages, and don't need the `--namespace-packages` flags
            # when being typechecked using mypy.
            with open(f"{self.python_folder}/feast/{sub_folder}/__init__.py", "w"):
                pass

        with open(f"{self.python_folder}/__init__.py", "w"):
            pass
        with open(f"{self.python_folder}/feast/__init__.py", "w"):
            pass

        for path in Path(self.python_folder).rglob("*.py"):
            for folder in self.sub_folders:
                # Read in the file
                with open(path, "r") as file:
                    filedata = file.read()

                # Replace the target string
                filedata = filedata.replace(
                    f"from feast.{folder}", f"from feast.protos.feast.{folder}"
                )

                # Write the file out again
                with open(path, "w") as file:
                    file.write(filedata)


class BuildCommand(build_py):
    """Custom build command."""

    def run(self):
        self.run_command("build_python_protos")

        self.run_command("build_ext")
        build_py.run(self)


class DevelopCommand(develop):
    """Custom develop command."""

    def run(self):
        self.reinitialize_command("build_python_protos", inplace=1)
        self.run_command("build_python_protos")

        develop.run(self)


setup(
    name=NAME,
    author=AUTHOR,
    description=DESCRIPTION,
    version=VERSION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        where=PYTHON_CODE_PREFIX, exclude=("java", "infra", "sdk/python/tests", "ui")
    ),
    package_dir={"": PYTHON_CODE_PREFIX},
    install_requires=REQUIRED,
    # https://stackoverflow.com/questions/28509965/setuptools-development-requirements
    # Install dev requirements with: pip install -e .[dev]
    extras_require={
        "dev": DEV_REQUIRED,
        "ci": CI_REQUIRED,
        "gcp": GCP_REQUIRED,
        "aws": AWS_REQUIRED,
        "bytewax": BYTEWAX_REQUIRED,
        "k8s": KUBERNETES_REQUIRED,
        "redis": REDIS_REQUIRED,
        "snowflake": SNOWFLAKE_REQUIRED,
        "spark": SPARK_REQUIRED,
        "trino": TRINO_REQUIRED,
        "postgres": POSTGRES_REQUIRED,
        "azure": AZURE_REQUIRED,
        "mysql": MYSQL_REQUIRED,
        "ge": GE_REQUIRED,
        "hbase": HBASE_REQUIRED,
        "docs": DOCS_REQUIRED,
        "cassandra": CASSANDRA_REQUIRED,
        "hazelcast": HAZELCAST_REQUIRED,
        "grpcio": GRPCIO_REQUIRED,
        "rockset": ROCKSET_REQUIRED,
        "ibis": IBIS_REQUIRED,
        "duckdb": DUCKDB_REQUIRED,
    },
    include_package_data=True,
    license="Apache",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={"console_scripts": ["feast=feast.cli:cli"]},
    use_scm_version=use_scm_version,
    setup_requires=[
        "setuptools_scm",
        "grpcio>=1.56.2,<2",
        "grpcio-tools>=1.56.2,<2",
        "mypy-protobuf>=3.1",
        "pybindgen==0.22.0",
    ],
    cmdclass={
        "build_python_protos": BuildPythonProtosCommand,
        "build_py": BuildCommand,
        "develop": DevelopCommand,
    },
)



import logging
from abc import abstractmethod
from datetime import datetime, timedelta
from threading import Lock
from typing import List, Optional

from feast import usage
from feast.data_source import DataSource
from feast.entity import Entity
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.infra.infra_object import Infra
from feast.infra.registry import proto_registry_utils
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.project_metadata import ProjectMetadata
from feast.saved_dataset import SavedDataset, ValidationReference
from feast.stream_feature_view import StreamFeatureView

logger = logging.getLogger(__name__)


class CachingRegistry(BaseRegistry):
    def __init__(
        self,
        project: str,
        cache_ttl_seconds: int,
    ):
        self.cached_registry_proto = self.proto()
        proto_registry_utils.init_project_metadata(self.cached_registry_proto, project)
        self.cached_registry_proto_created = datetime.utcnow()
        self._refresh_lock = Lock()
        self.cached_registry_proto_ttl = timedelta(
            seconds=cache_ttl_seconds if cache_ttl_seconds is not None else 0
        )

    @abstractmethod
    def _get_data_source(self, name: str, project: str) -> DataSource:
        pass

    def get_data_source(
        self, name: str, project: str, allow_cache: bool = False
    ) -> DataSource:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_data_source(
                self.cached_registry_proto, name, project
            )
        return self._get_data_source(name, project)

    @abstractmethod
    def _list_data_sources(self, project: str) -> List[DataSource]:
        pass

    def list_data_sources(
        self, project: str, allow_cache: bool = False
    ) -> List[DataSource]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_data_sources(
                self.cached_registry_proto, project
            )
        return self._list_data_sources(project)

    @abstractmethod
    def _get_entity(self, name: str, project: str) -> Entity:
        pass

    def get_entity(self, name: str, project: str, allow_cache: bool = False) -> Entity:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_entity(
                self.cached_registry_proto, name, project
            )
        return self._get_entity(name, project)

    @abstractmethod
    def _list_entities(self, project: str) -> List[Entity]:
        pass

    def list_entities(self, project: str, allow_cache: bool = False) -> List[Entity]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_entities(
                self.cached_registry_proto, project
            )
        return self._list_entities(project)

    @abstractmethod
    def _get_feature_view(self, name: str, project: str) -> FeatureView:
        pass

    def get_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureView:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_feature_view(name, project)

    @abstractmethod
    def _list_feature_views(self, project: str) -> List[FeatureView]:
        pass

    def list_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[FeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_feature_views(project)

    @abstractmethod
    def _get_on_demand_feature_view(
        self, name: str, project: str
    ) -> OnDemandFeatureView:
        pass

    def get_on_demand_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> OnDemandFeatureView:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_on_demand_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_on_demand_feature_view(name, project)

    @abstractmethod
    def _list_on_demand_feature_views(self, project: str) -> List[OnDemandFeatureView]:
        pass

    def list_on_demand_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[OnDemandFeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_on_demand_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_on_demand_feature_views(project)

    @abstractmethod
    def _get_stream_feature_view(self, name: str, project: str) -> StreamFeatureView:
        pass

    def get_stream_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> StreamFeatureView:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_stream_feature_view(
                self.cached_registry_proto, name, project
            )
        return self._get_stream_feature_view(name, project)

    @abstractmethod
    def _list_stream_feature_views(self, project: str) -> List[StreamFeatureView]:
        pass

    def list_stream_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[StreamFeatureView]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_stream_feature_views(
                self.cached_registry_proto, project
            )
        return self._list_stream_feature_views(project)

    @abstractmethod
    def _get_feature_service(self, name: str, project: str) -> FeatureService:
        pass

    def get_feature_service(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureService:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_feature_service(
                self.cached_registry_proto, name, project
            )
        return self._get_feature_service(name, project)

    @abstractmethod
    def _list_feature_services(self, project: str) -> List[FeatureService]:
        pass

    def list_feature_services(
        self, project: str, allow_cache: bool = False
    ) -> List[FeatureService]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_feature_services(
                self.cached_registry_proto, project
            )
        return self._list_feature_services(project)

    @abstractmethod
    def _get_saved_dataset(self, name: str, project: str) -> SavedDataset:
        pass

    def get_saved_dataset(
        self, name: str, project: str, allow_cache: bool = False
    ) -> SavedDataset:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_saved_dataset(
                self.cached_registry_proto, name, project
            )
        return self._get_saved_dataset(name, project)

    @abstractmethod
    def _list_saved_datasets(self, project: str) -> List[SavedDataset]:
        pass

    def list_saved_datasets(
        self, project: str, allow_cache: bool = False
    ) -> List[SavedDataset]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_saved_datasets(
                self.cached_registry_proto, project
            )
        return self._list_saved_datasets(project)

    @abstractmethod
    def _get_validation_reference(self, name: str, project: str) -> ValidationReference:
        pass

    def get_validation_reference(
        self, name: str, project: str, allow_cache: bool = False
    ) -> ValidationReference:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.get_validation_reference(
                self.cached_registry_proto, name, project
            )
        return self._get_validation_reference(name, project)

    @abstractmethod
    def _list_validation_references(self, project: str) -> List[ValidationReference]:
        pass

    def list_validation_references(
        self, project: str, allow_cache: bool = False
    ) -> List[ValidationReference]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_validation_references(
                self.cached_registry_proto, project
            )
        return self._list_validation_references(project)

    @abstractmethod
    def _list_project_metadata(self, project: str) -> List[ProjectMetadata]:
        pass

    def list_project_metadata(
        self, project: str, allow_cache: bool = False
    ) -> List[ProjectMetadata]:
        if allow_cache:
            self._refresh_cached_registry_if_necessary()
            return proto_registry_utils.list_project_metadata(
                self.cached_registry_proto, project
            )
        return self._list_project_metadata(project)

    @abstractmethod
    def _get_infra(self, project: str) -> Infra:
        pass

    def get_infra(self, project: str, allow_cache: bool = False) -> Infra:
        return self._get_infra(project)

    def refresh(self, project: Optional[str] = None):
        if project:
            project_metadata = proto_registry_utils.get_project_metadata(
                registry_proto=self.cached_registry_proto, project=project
            )
            if project_metadata:
                usage.set_current_project_uuid(project_metadata.project_uuid)
            else:
                proto_registry_utils.init_project_metadata(
                    self.cached_registry_proto, project
                )
        self.cached_registry_proto = self.proto()
        self.cached_registry_proto_created = datetime.utcnow()

    def _refresh_cached_registry_if_necessary(self):
        with self._refresh_lock:
            expired = (
                self.cached_registry_proto is None
                or self.cached_registry_proto_created is None
            ) or (
                self.cached_registry_proto_ttl.total_seconds()
                > 0  # 0 ttl means infinity
                and (
                    datetime.utcnow()
                    > (
                        self.cached_registry_proto_created
                        + self.cached_registry_proto_ttl
                    )
                )
            )

            if expired:
                logger.info("Registry cache expired, so refreshing")
                self.refresh()



import json
from typing import Callable, Dict, Iterable, Optional, Tuple

import pandas as pd
from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType
from feast.infra.utils.doris.connection_util import _get_conn
from feast.infra.utils.doris.doris_config import doris_type_to_feast_value_type
from typeguard import typechecked


@typechecked
class DorisSQLSource(DataSource):
    def __init__(
        self,
        name: Optional[str] = None,
        query: Optional[str] = None,
        table: Optional[str] = None,
        timestamp_field: Optional[str] = "",
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
    ):
        self._doris_options = DorisSQLOptions(name=name, query=query, table=table)
        if name is None and table is None:
            raise DataSourceNoNameException()
        name = name or table
        assert name

        super().__init__(
            name=name,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            description=description,
            tags=tags,
            owner=owner,
        )

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, DorisSQLSource):
            raise TypeError(
                "Comparisons should only involve DorisSQLSource class objects."
            )

        return (
            super().__eq__(other)
            and self._doris_options._query == other._doris_options._query
            and self.timestamp_field == other.timestamp_field
            and self.created_timestamp_column == other.created_timestamp_column
            and self.field_mapping == other.field_mapping
        )

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        assert data_source.HasField("custom_options")

        doris_options = json.loads(data_source.custom_options.configuration)

        return DorisSQLSource(
            name=doris_options["name"],
            query=doris_options["query"],
            table=doris_options["table"],
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
        )

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            # Change this to doris path
            data_source_class_type="feast.infra.offline_stores.contrib.doris_offline_store.doris_source.DorisSQLSource",
            field_mapping=self.field_mapping,
            custom_options=self._doris_options.to_proto(),
            description=self.description,
            tags=self.tags,
            owner=self.owner,
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        pass

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return doris_type_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        with _get_conn(config.offline_store) as conn, conn.cursor() as cur:
            sql_query = f"SELECT * FROM {self.get_table_query_string()} AS sub LIMIT 0"
            df = pd.read_sql(
                sql_query,
                conn,
            )
            columns = list(df.columns)
            column_list = "', '".join(columns)
            query = f"""
            SELECT
                COLUMN_NAME,
                DATA_TYPE
            FROM
                information_schema.COLUMNS
            WHERE
                TABLE_SCHEMA = '{config.offline_store.database}'
                AND COLUMN_NAME IN ('{column_list}');
            """
            cur.execute(query)
            data = cur.fetchall()
            return data

    def get_table_query_string(self) -> str:
        if self._doris_options._table:
            return f"{self._doris_options._table}"
        else:
            return f"({self._doris_options._query})"


class DorisSQLOptions:
    def __init__(
        self,
        name: Optional[str],
        query: Optional[str],
        table: Optional[str],
    ):
        self._name = name or ""
        self._query = query or ""
        self._table = table or ""

    @classmethod
    def from_proto(cls, doris_options_proto: DataSourceProto.CustomSourceOptions):
        config = json.loads(doris_options_proto.configuration.decode("utf8"))
        doris_options = cls(
            name=config["name"], query=config["query"], table=config["table"]
        )

        return doris_options

    def to_proto(self) -> DataSourceProto.CustomSourceOptions:
        doris_options_proto = DataSourceProto.CustomSourceOptions(
            configuration=json.dumps(
                {"name": self._name, "query": self._query, "table": self._table}
            ).encode()
        )
        return doris_options_proto


class SavedDatasetDorisSQLStorage(SavedDatasetStorage):
    _proto_attr_name = "custom_storage"

    doris_options: DorisSQLOptions

    def __init__(self, table_ref: str):
        self.doris_options = DorisSQLOptions(table=table_ref, name=None, query=None)

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        return SavedDatasetDorisSQLStorage(
            table_ref=DorisSQLOptions.from_proto(storage_proto.custom_storage)._table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(custom_storage=self.doris_options.to_proto())

    def to_data_source(self) -> DataSource:
        return DorisSQLSource(table=self.doris_options._table)



import calendar
import struct
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

from happybase import ConnectionPool
from happybase.connection import DEFAULT_PROTOCOL, DEFAULT_TRANSPORT
from pydantic import StrictStr

from feast import Entity
from feast.feature_view import FeatureView
from feast.infra.online_stores.helpers import compute_entity_id
from feast.infra.online_stores.online_store import OnlineStore
from feast.infra.utils.hbase_utils import HBaseConnector, HbaseConstants
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.usage import log_exceptions_and_usage


class HbaseOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for Hbase store"""

    type: Literal["hbase"] = "hbase"
    """Online store type selector"""

    host: StrictStr
    """Hostname of Hbase Thrift server"""

    port: StrictStr
    """Port in which Hbase Thrift server is running"""

    connection_pool_size: int = 4
    """Number of connections to Hbase Thrift server to keep in the connection pool"""

    protocol: StrictStr = DEFAULT_PROTOCOL
    """Protocol used to communicate with Hbase Thrift server"""

    transport: StrictStr = DEFAULT_TRANSPORT
    """Transport used to communicate with Hbase Thrift server"""


class HbaseOnlineStore(OnlineStore):
    """
    Online feature store for Hbase.

    Attributes:
        _conn: Happybase Connection to connect to hbase thrift server.
    """

    _conn: ConnectionPool = None

    def _get_conn(self, config: RepoConfig):
        """
        Get or Create Hbase Connection from Repoconfig.

        Args:
            config: The RepoConfig for the current FeatureStore.
        """

        store_config = config.online_store
        assert isinstance(store_config, HbaseOnlineStoreConfig)

        if not self._conn:
            self._conn = ConnectionPool(
                host=store_config.host,
                port=int(store_config.port),
                size=int(store_config.connection_pool_size),
                protocol=store_config.protocol,
                transport=store_config.transport,
            )
        return self._conn

    @log_exceptions_and_usage(online_store="hbase")
    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Write a batch of feature rows to Hbase online store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            data: a list of quadruplets containing Feature data. Each quadruplet contains an Entity Key,
            a dict containing feature values, an event timestamp for the row, and
            the created timestamp for the row if it exists.
            progress: Optional function to be called once every mini-batch of rows is written to
            the online store. Can be used to display progress.
        """

        hbase = HBaseConnector(self._get_conn(config))
        project = config.project
        table_name = self._table_id(project, table)

        b = hbase.batch(table_name)
        for entity_key, values, timestamp, created_ts in data:
            row_key = self._hbase_row_key(
                entity_key,
                feature_view_name=table.name,
                config=config,
            )
            values_dict = {}
            for feature_name, val in values.items():
                values_dict[HbaseConstants.get_col_from_feature(feature_name)] = (
                    val.SerializeToString()
                )
            if isinstance(timestamp, datetime):
                values_dict[HbaseConstants.DEFAULT_EVENT_TS] = struct.pack(
                    ">L", int(calendar.timegm(timestamp.timetuple()))
                )
            else:
                values_dict[HbaseConstants.DEFAULT_EVENT_TS] = timestamp
            if created_ts is not None:
                if isinstance(created_ts, datetime):
                    values_dict[HbaseConstants.DEFAULT_CREATED_TS] = struct.pack(
                        ">L", int(calendar.timegm(created_ts.timetuple()))
                    )
                else:
                    values_dict[HbaseConstants.DEFAULT_CREATED_TS] = created_ts
            b.put(row_key, values_dict)
        b.send()

        if progress:
            progress(len(data))

    @log_exceptions_and_usage(online_store="hbase")
    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Retrieve feature values from the Hbase online store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            entity_keys: a list of entity keys that should be read from the FeatureStore.
            requested_features: a list of requested feature names.
        """
        hbase = HBaseConnector(self._get_conn(config))
        project = config.project
        table_name = self._table_id(project, table)

        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []

        row_keys = [
            self._hbase_row_key(
                entity_key,
                feature_view_name=table.name,
                config=config,
            )
            for entity_key in entity_keys
        ]
        rows = hbase.rows(table_name, row_keys=row_keys)

        for _, row in rows:
            res = {}
            res_ts = None
            for feature_name, feature_value in row.items():
                f_name = HbaseConstants.get_feature_from_col(feature_name)
                if requested_features is not None and f_name in requested_features:
                    v = ValueProto()
                    v.ParseFromString(feature_value)
                    res[f_name] = v
                if f_name is HbaseConstants.EVENT_TS:
                    ts = struct.unpack(">L", feature_value)[0]
                    res_ts = datetime.fromtimestamp(ts)
            if not res:
                result.append((None, None))
            else:
                result.append((res_ts, res))
        return result

    @log_exceptions_and_usage(online_store="hbase")
    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        Update tables from the Hbase Online Store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables_to_delete: Tables to delete from the Hbase Online Store.
            tables_to_keep: Tables to keep in the Hbase Online Store.
        """
        hbase = HBaseConnector(self._get_conn(config))
        project = config.project

        # We don't create any special state for the entites in this implementation.
        for table in tables_to_keep:
            table_name = self._table_id(project, table)
            if not hbase.check_if_table_exist(table_name):
                hbase.create_table_with_default_cf(table_name)

        for table in tables_to_delete:
            table_name = self._table_id(project, table)
            hbase.delete_table(table_name)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        """
        Delete tables from the Hbase Online Store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables: Tables to delete from the feature repo.
        """
        hbase = HBaseConnector(self._get_conn(config))
        project = config.project

        for table in tables:
            table_name = self._table_id(project, table)
            hbase.delete_table(table_name)

    def _hbase_row_key(
        self,
        entity_key: EntityKeyProto,
        feature_view_name: str,
        config: RepoConfig,
    ) -> bytes:
        """
        Computes the HBase row key for a given entity key and feature view name.

        Args:
            entity_key (EntityKeyProto): The entity key to compute the row key for.
            feature_view_name (str): The name of the feature view to compute the row key for.
            config (RepoConfig): The configuration for the Feast repository.

        Returns:
            bytes: The HBase row key for the given entity key and feature view name.
        """
        entity_id = compute_entity_id(
            entity_key,
            entity_key_serialization_version=config.entity_key_serialization_version,
        )
        # Even though `entity_id` uniquely identifies an entity, we use the same table
        # for multiple feature_views with the same set of entities.
        # To uniquely identify the row for a feature_view, we suffix the name of the feature_view itself.
        # This also ensures that features for entities from various feature_views are
        # colocated.
        return f"{entity_id}#{feature_view_name}".encode()

    def _table_id(self, project: str, table: FeatureView) -> str:
        """
        Returns table name given the project_name and the feature_view.

        Args:
            project: Name of the feast project.
            table: Feast FeatureView.
        """
        return f"{project}:{table.name}"



from __future__ import absolute_import

from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

import pymysql
import pytz
from pydantic import StrictStr
from pymysql.connections import Connection
from pymysql.cursors import Cursor

from feast import Entity, FeatureView, RepoConfig
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel


class MySQLOnlineStoreConfig(FeastConfigBaseModel):
    """
    Configuration for the MySQL online store.
    NOTE: The class *must* end with the `OnlineStoreConfig` suffix.
    """

    type: Literal["mysql"] = "mysql"

    host: Optional[StrictStr] = None
    user: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    database: Optional[StrictStr] = None
    port: Optional[int] = None


class MySQLOnlineStore(OnlineStore):
    """
    An online store implementation that uses MySQL.
    NOTE: The class *must* end with the `OnlineStore` suffix.
    """

    _conn: Optional[Connection] = None

    def _get_conn(self, config: RepoConfig) -> Connection:
        online_store_config = config.online_store
        assert isinstance(online_store_config, MySQLOnlineStoreConfig)

        if not self._conn:
            self._conn = pymysql.connect(
                host=online_store_config.host or "127.0.0.1",
                user=online_store_config.user or "test",
                password=online_store_config.password or "test",
                database=online_store_config.database or "feast",
                port=online_store_config.port or 3306,
                autocommit=True,
            )
        return self._conn

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        conn = self._get_conn(config)
        cur = conn.cursor()

        project = config.project

        for entity_key, values, timestamp, created_ts in data:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=2,
            ).hex()
            timestamp = _to_naive_utc(timestamp)
            if created_ts is not None:
                created_ts = _to_naive_utc(created_ts)

            for feature_name, val in values.items():
                self.write_to_table(
                    created_ts,
                    cur,
                    entity_key_bin,
                    feature_name,
                    project,
                    table,
                    timestamp,
                    val,
                )
            conn.commit()
            if progress:
                progress(1)

    @staticmethod
    def write_to_table(
        created_ts, cur, entity_key_bin, feature_name, project, table, timestamp, val
    ) -> None:
        cur.execute(
            f"""
            INSERT INTO {_table_id(project, table)}
            (entity_key, feature_name, value, event_ts, created_ts)
            values (%s, %s, %s, %s, %s)
            ON DUPLICATE KEY UPDATE
            value = %s,
            event_ts = %s,
            created_ts = %s;
            """,
            (
                # Insert
                entity_key_bin,
                feature_name,
                val.SerializeToString(),
                timestamp,
                created_ts,
                # Update on duplicate key
                val.SerializeToString(),
                timestamp,
                created_ts,
            ),
        )

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        conn = self._get_conn(config)
        cur = conn.cursor()

        result: List[Tuple[Optional[datetime], Optional[Dict[str, Any]]]] = []

        project = config.project
        for entity_key in entity_keys:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=2,
            ).hex()

            cur.execute(
                f"SELECT feature_name, value, event_ts FROM {_table_id(project, table)} WHERE entity_key = %s",
                (entity_key_bin,),
            )

            res = {}
            res_ts: Optional[datetime] = None
            records = cur.fetchall()
            if records:
                for feature_name, val_bin, ts in records:
                    val = ValueProto()
                    val.ParseFromString(val_bin)
                    res[feature_name] = val
                    res_ts = ts

            if not res:
                result.append((None, None))
            else:
                result.append((res_ts, res))
        return result

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ) -> None:
        conn = self._get_conn(config)
        cur = conn.cursor()
        project = config.project

        # We don't create any special state for the entities in this implementation.
        for table in tables_to_keep:
            table_name = _table_id(project, table)
            index_name = f"{table_name}_ek"
            cur.execute(
                f"""CREATE TABLE IF NOT EXISTS {table_name} (entity_key VARCHAR(512),
                feature_name VARCHAR(256),
                value BLOB,
                event_ts timestamp NULL DEFAULT NULL,
                created_ts timestamp NULL DEFAULT NULL,
                PRIMARY KEY(entity_key, feature_name))"""
            )

            index_exists = cur.execute(
                f"""
                SELECT 1 FROM information_schema.statistics
                WHERE table_schema = DATABASE() AND table_name = '{table_name}' AND index_name = '{index_name}'
                """
            )
            if not index_exists:
                cur.execute(
                    f"ALTER TABLE {table_name} ADD INDEX {index_name} (entity_key);"
                )

        for table in tables_to_delete:
            _drop_table_and_index(cur, project, table)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ) -> None:
        conn = self._get_conn(config)
        cur = conn.cursor()
        project = config.project

        for table in tables:
            _drop_table_and_index(cur, project, table)


def _drop_table_and_index(cur: Cursor, project: str, table: FeatureView) -> None:
    table_name = _table_id(project, table)
    cur.execute(f"DROP INDEX {table_name}_ek ON {table_name};")
    cur.execute(f"DROP TABLE IF EXISTS {table_name}")


def _table_id(project: str, table: FeatureView) -> str:
    return f"{project}_{table.name}"


def _to_naive_utc(ts: datetime) -> datetime:
    if ts.tzinfo is None:
        return ts
    else:
        return ts.astimezone(pytz.utc).replace(tzinfo=None)
